"""
Helper classes for managing feature layers and datasets.  These class are not created by users directly.
Instances of this class, are available as a properties of feature layers and make it easier to manage them.
"""
from __future__ import absolute_import
import os
import json
import time
import tempfile
import collections

from arcgis._impl.common._mixins import PropertyMap
from arcgis.gis import _GISResource

# pylint: disable=protected-access

class AttachmentManager(object):
    """
    Manager class for manipulating feature layer attachments. This class is not created by users directly.
    An instance of this class, called 'attachments', is available as a property of the FeatureLayer object,
    if the layer supports attachments.
    Users call methods on this 'attachments' object to manipulate (create, get, list, delete) attachments.
    """

    def __init__(self, layer):
        self._layer = layer

    def search(self, where='1=1', object_ids=None, global_ids=None,
              attachment_types=None, size=None, keywords=None,
              show_images=False, as_df=False):
        """

        The `search` method allows querying the layer for its attachments and returns the results as
        a Pandas DataFrame or dict


        =========================   ===============================================================
        **Arguement**               **Description**
        -------------------------   ---------------------------------------------------------------
        where                       required string.  The definition expression to be applied to
                                    the related layer/table. From the list of records that are
                                    related to the specified object Ids, only those records that
                                    conform to this expression will be returned.

                                    Example: where="STATE_NAME = 'Alaska'".
                                    The query results will return all attachments in Alaska.
        -------------------------   ---------------------------------------------------------------
        object_ids                  optional list/string. The object IDs of this layer/table to be
                                    queried.

                                    Syntax: objectIds=<objectId1>,<objectId2>

                                    Example: objectIds=2. The query results will return attachments
                                    only for the specified object id.
        -------------------------   ---------------------------------------------------------------
        global_ids                   optional list/string. The global IDs of this layer/table to be
                                    queried.

                                    Syntax: globalIds=<globalIds1>,<globalIds2>

                                    Example: globalIds=6s430c5a-kb75-4d52-a0db-b30bg060f0b9,35f0d027-8fc0-4905-a2f6-373c9600d017

                                    The query results will return attachments only for specified
                                    global id.
        -------------------------   ---------------------------------------------------------------
        attachment_types            optional list/string. The file format that is supported by
                                    query attachment.

                                    Supported attachment types:
                                    bmp, ecw, emf, eps, ps, gif, img, jp2, jpc, j2k, jpf, jpg,
                                    jpeg, jpe, png, psd, raw, sid, tif, tiff, wmf, wps, avi, mpg,
                                    mpe, mpeg, mov, wmv, aif, mid, rmi, mp2, mp3, mp4, pma, mpv2,
                                    qt, ra, ram, wav, wma, doc, docx, dot, xls, xlsx, xlt, pdf, ppt,
                                    pptx, txt, zip, 7z, gz, gtar, tar, tgz, vrml, gml, json, xml,
                                    mdb, geodatabase

                                    Example: attachment_types='image/jpeg'
        -------------------------   ---------------------------------------------------------------
        size                        optional tuple/list. The file size of the attachment is
                                    specified in bytes. You can enter a file size range
                                    (1000,15000) to query for attachments with the specified range.

                                    Example: size=1000,15000.
                                    The query results will return all attachments within the
                                    specified file size range (1000 - 15000) bytes.
        -------------------------   ---------------------------------------------------------------
        keywords                    optional string.  When attachments are uploaded, keywords can
                                    be assigned to the uploaded file.  By passing a keyword value,
                                    the values will be searched.

                                    Example: keywords='airplanes'
        -------------------------   ---------------------------------------------------------------
        show_images                 optional bool. The default is False, when the value is True,
                                    the results will be displayed as a HTML table. If the as_df is
                                    set to False, this parameter will be ignored.
        -------------------------   ---------------------------------------------------------------
        as_df                       optional bool. Default is False, if True, the results will be
                                    a Pandas' DataFrame.  If False, the values will be a list of
                                    dictionary values.
        =========================   ===============================================================

        :returns: list of downloaded files

        """
        import copy
        columns = [col.upper() for col in ['ParentObjectid', 'ParentGlobalId', 'Id',
                                           'Name', 'GlobalId', 'ContentType',
                                           'Size','KeyWords', 'URL', 'IMAGE_PREVIEW']]
        result_offset = 0
        if keywords is None:
            keywords = []
        elif isinstance(keywords, str):
            keywords = keywords.split(',')
        if object_ids is None:
            object_ids = []
        elif isinstance(object_ids, str):
            object_ids = object_ids.split(',')
        if global_ids is None:
            global_ids = []
        elif isinstance(global_ids, str):
            global_ids = global_ids.split(',')
        if attachment_types is None:
            attachment_types = []
        elif isinstance(attachment_types, str):
            attachment_types = attachment_types.split(',')
        if isinstance(size, (tuple, list)):
            size = ",".join(list([str(s) for s in size]))
        elif size is None:
            size = None
        if self._layer._gis._portal.is_arcgisonline == False and \
           self._layer.properties.hasAttachments:
            rows = []

            query = self._layer.query(where=where,
                                      object_ids=",".join(object_ids),
                                      global_ids=",".join(global_ids),
                                      return_ids_only=True)
            if 'objectIds' in query:
                token = self._layer._con.token
                for i in query['objectIds']:
                    attachments = self.get_list(oid=i)
                    for att in attachments:
                        if not token is None:
                            att_path = '{}/{}/attachments/{}?token={}'.format(self._layer.url, i, att['id'], self._layer._con.token)
                        else:
                            att_path = '{}/{}/attachments/{}'.format(self._layer.url, i, att['id'])
                        preview = None
                        if att['contentType'].find("image") > -1:
                            preview = "<img src=\"" + att_path + "\" width=150 height=150 />"

                        row = {
                            "PARENTOBJECTID" : i,
                            "PARENTGLOBALID" : "N/A",
                            "ID" : att['id'],
                            "NAME" : att['name'],
                            "CONTENTTYPE" : att['contentType'],
                            "SIZE" : att['size'],
                            "KEYWORDS" : "",
                            "IMAGE_PREVIEW" : preview
                        }
                        if 'globalId' in att:
                            row["GLOBALID"] = att['globalId']
                        if as_df and show_images:
                            row["DOWNLOAD_URL"] = "<a href=\"%s\" target=\"_blank\">DATA</a>" % att_path
                        else:
                            row["DOWNLOAD_URL"] = "%s" % att_path
                        rows.append(row)

                if attachment_types is not None and \
                   len(attachment_types) > 0: # performs contenttype search
                    if isinstance(attachment_types, str):
                        attachment_types= attachment_types.split(',')
                    rows = [row for row in rows \
                            if os.path.splitext(row['NAME'])[1][1:] in attachment_types or \
                            row['CONTENTTYPE'] in attachment_types]
        else:
            url = '{}/{}'.format(self._layer.url, 'queryAttachments')
            params = {
                'f' : 'json',
                "attachmentTypes" : ",".join(attachment_types),
                'objectIds' : ",".join([str(v) for v in object_ids]),
                'globalIds' : ",".join([str(v) for v in global_ids]),
                "definitionExpression" : where,
                'keywords' : ",".join([str(v) for v in keywords]),
                'size' : None,
            }
            iterparams = copy.copy(params)
            for k,v in iterparams.items():
                if k in ['objectIds', 'globalIds',
                         'attachmentTypes'] and \
                   v == "":
                    del params[k]
                elif k == 'size' and \
                     v is None:
                    del params[k]

            results = self._layer._con.post(url,params)
            rows = []
            if 'attachmentGroups' not in results:
                return []
            for result in results['attachmentGroups']:
                for data in result['attachmentInfos']:
                    token = self._layer._con.token
                    if not token is None:
                        att_path = '{}/{}/attachments/{}?token={}'.format(self._layer.url, result['parentObjectId'], data['id'], self._layer._con.token)
                    else:
                        att_path = '{}/{}/attachments/{}'.format(self._layer.url, result['parentObjectId'], data['id'])
                    preview = None
                    if data['contentType'].find("image") > -1:
                        preview = "<img src=\"" + att_path + "\" width=150 height=150 />"

                    row = {
                        "PARENTOBJECTID" : result['parentObjectId'],
                        "PARENTGLOBALID" : result['parentGlobalId'],
                        "ID" : data['id'],
                        "NAME" : data['name'],
                        "CONTENTTYPE" : data['contentType'],
                        "SIZE" : data['size'],
                        "KEYWORDS" : data['keywords'],
                        "IMAGE_PREVIEW" : preview
                    }
                    if 'globalId' in data:
                        row["GLOBALID"] = data['globalId']
                    if as_df and show_images:
                        row["DOWNLOAD_URL"] = "<a href=\"%s\" target=\"_blank\">DATA</a>" % att_path
                    else:
                        row["DOWNLOAD_URL"] = "%s" % att_path
                    rows.append(row)
                    del row

        if as_df == True:
            import pandas as pd
            if show_images:
                from IPython.display import HTML
                pd.set_option('display.max_colwidth', -1)
                return HTML(pd.DataFrame.from_dict(rows).to_html(escape=False))
            else:
                df = pd.DataFrame.from_dict(rows)
                df.drop(['DOWNLOAD_URL', "IMAGE_PREVIEW"], axis=1, inplace=True)
                return df
        else:
            return rows
        return None

    def _download_all(self, object_ids=None, save_folder=None, attachment_types=None):
        """
        downloads all attachments to a specific folder

        =========================   ===============================================================
        **Arguement**               **Description**
        -------------------------   ---------------------------------------------------------------
        object_ids                  optional list. A list of object_ids to download data from.
        -------------------------   ---------------------------------------------------------------
        save_folder                 optional string. Path to save data to.
        -------------------------   ---------------------------------------------------------------
        attachment_types            optional string.  Allows the limitation of file types by passing
                                    a string of the item type.

                                    **Example:** image/jpeg
        =========================   ===============================================================

        :returns: list of downloaded files

        """
        results = []
        if save_folder is None:
            save_folder = os.path.join(tempfile.gettempdir(), 'attachment_download')
        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)
        attachments = self.search(object_ids=object_ids, attachment_types=attachment_types, as_df=True)
        for row in attachments.to_dict(orient='records') :
            dlpath = os.path.join(save_folder, "%s" % int(row['PARENTOBJECTID']), "%s" % int(row['ID']))
            if os.path.isdir(dlpath) == False:
                os.makedirs(dlpath)
            path = self.download(oid=int(row['PARENTOBJECTID']), attachment_id=int(row['ID']), save_path=dlpath)
            results.append(path[0])
            del row
        return results

    def get_list(self, oid):
        """ returns the list of attachements for a given OBJECT ID """
        return self._layer._list_attachments(oid)['attachmentInfos']

    def download(self, oid=None, attachment_id=None, save_path=None):
        """
        downloads attachment and returns it's path on disk.

        The download tool works as follows:

            1). if nothing is given, all attachments will be downloaded
               - example: download()
            2). If a single oid and attachment_id are given, the single file will download
            3). If a list of oid values are given, all the attachments for those object ids will be saved locally.



        =========================   ===============================================================
        **Arguement**               **Description**
        -------------------------   ---------------------------------------------------------------
        oid                         optional list/string. A list of object Ids or a single value
                                    to download data from.
        -------------------------   ---------------------------------------------------------------
        attachment_id               optional string. Id of the attachment to download. This is only
                                    honored if return_all is False.
        -------------------------   ---------------------------------------------------------------
        save_folder                 optional string. Path to save data to.
        =========================   ===============================================================

        :returns: list of downloaded files


        """
        return_all = False
        if isinstance(oid, str):
            oid = oid.split(',')
            oid_len = len(oid)
        elif isinstance(oid, int):
            oid = str(oid).split(',')
            oid_len = 1
        elif isinstance(oid, (tuple, list)):
            oid_len = len(oid)
        elif oid is None:
            oid_len = 0
        else:
            raise ValueError("oid must be of type list or string")
        if isinstance(attachment_id, str):
            attachment_id = [int(att) for att in attachment_id.split(',')]
            att_len = len(attachment_id)
        elif isinstance(attachment_id, int):
            attachment_id = str(attachment_id).split(',')
            att_len = 1
        elif isinstance(attachment_id, (tuple, list)):
            att_len = len(attachment_id)
        elif attachment_id is None:
            att_len = 0
        else:
            raise ValueError("attachment_id must be of type list or string")
        if oid_len == 1 and att_len > 0:
            return_all = False
        elif oid_len > 1 and att_len > 0:
            raise ValueError("You cannot provide more than one oid when providing attachment_id values.")
        else:
            return_all = True

        if not return_all:
            oid = oid[0]
            paths = []
            for att in attachment_id:
                att_path = '{}/{}/attachments/{}'.format(self._layer.url, oid, att)
                att_list = self.get_list(int(oid))

                #get attachment file name
                desired_att = [att2 for att2 in att_list if att2['id'] == int(att)]
                if len(desired_att) == 0: #bad attachment id
                    raise RuntimeError
                else:
                    att_name = desired_att[0]['name']

                if not save_path:
                    save_path = tempfile.gettempdir()
                if not os.path.isdir(save_path):
                    os.makedirs(save_path)

                path = self._layer._con.get(path=att_path, try_json=False, out_folder=save_path,
                                            file_name=att_name, token=self._layer._token, force_bytes=False)
                paths.append(path)
            return paths
        else:
            return self._download_all(object_ids=oid,
                                      save_folder=save_path)

    def add(self, oid, file_path):
        """ Adds an attachment to a feature layer
            Input:
              oid - string - OBJECTID value to add attachment to
              file_path - string - path to file
            Output:
              JSON Repsonse
        """
        return self._layer._add_attachment(oid, file_path)

    def delete(self, oid, attachment_id):
        """ removes an attachment from a feature
            Input:
              oid - integer or string - id of feature
              attachment_id - integer - id of attachment to erase
            Output:
               JSON response
        """
        return self._layer._delete_attachment(oid, attachment_id)

    def update(self, oid, attachment_id, file_path):
        """ updates an existing attachment with a new file
            Inputs:
               oid - string/integer - Unique record ID
               attachment_id - integer - Unique attachment identifier
               file_path - string - path to new attachment
            Output:
               JSON response
        """
        return self._layer._update_attachment(oid, attachment_id, file_path)


class SyncManager(object):
    """
    Manager class for manipulating replicas for syncing disconnected editing of feature layers.
    This class is not created by users directly.
    An instance of this class, called 'replicas', is available as a property of the FeatureLayerCollection object,
    if the layer is sync enabled / supports disconnected editing.
    Users call methods on this 'replicas' object to manipulate (create, synchronize, unregister) replicas.
    """

    # http://services.arcgis.com/help/fsDisconnectedEditing.html
    def __init__(self, featsvc):
        self._fs = featsvc

    def get_list(self):
        """ returns all the replicas for the feature layer collection """
        return self._fs._replicas

    # ----------------------------------------------------------------------
    def unregister(self, replica_id):
        """
           unregisters a replica from a feature layer collection
           Inputs:
             replica_id - The replicaID returned by the feature service
                          when the replica was created.
        """
        return self._fs._unregister_replica(replica_id)

    # ----------------------------------------------------------------------
    def get(self, replica_id):
        """
           returns replica metadata for a specific replica.
           Inputs:
              replica_id - The replicaID returned by the feature service
                           when the replica was created.
        """
        return self._fs._replica_info(replica_id)

    # ----------------------------------------------------------------------
    def create(self,
               replica_name,
               layers,
               layer_queries=None,
               geometry_filter=None,
               replica_sr=None,
               transport_type="esriTransportTypeUrl",
               return_attachments=False,
               return_attachments_databy_url=False,
               asynchronous=False,
               attachments_sync_direction="none",
               sync_model="none",
               data_format="json",
               replica_options=None,
               wait=False,
               out_path=None,
               sync_direction=None,
               target_type="client"):
        """
        The create operation is performed on a feature layer collection
        resource. This operation creates the replica between the feature
        dataset and a client based on a client-supplied replica definition.
        It requires the Sync capability. See Sync overview for more
        information on sync. The response for create includes
        replicaID, replica generation number, and data similar to the
        response from the feature layer collection query operation.
        The create operation returns a response of type
        esriReplicaResponseTypeData, as the response has data for the
        layers in the replica. If the operation is called to register
        existing data by using replicaOptions, the response type will be
        esriReplicaResponseTypeInfo, and the response will not contain data
        for the layers in the replica.

        Inputs:
           replica_name - name of the replica
           layers - layers to export
           layer_queries - In addition to the layers and geometry parameters, the layer_queries
            parameter can be used to further define what is replicated. This
            parameter allows you to set properties on a per layer or per table
            basis. Only the properties for the layers and tables that you want
            changed from the default are required.
            Example:
             layer_queries = {"0":{"queryOption": "useFilter", "useGeometry": true,
             "where": "requires_inspection = Yes"}}
           geometry_filter - spatial filter from arcgis.geometry.filters module to filter results by a
                             spatial relationship with another geometry
           replica_sr - the spatial reference of the replica geometry.
           transport_type -  The transport_type represents the response format. If the
            transport_type is esriTransportTypeUrl, the JSON response is contained in a file,
            and the URL link to the file is returned. Otherwise, the JSON object is returned
            directly. The default is esriTransportTypeUrl.
            If async is true, the results will always be returned as if transport_type is
            esriTransportTypeUrl. If dataFormat is sqlite, the transportFormat will always be
            esriTransportTypeUrl regardless of how the parameter is set.
            Values: esriTransportTypeUrl | esriTransportTypeEmbedded
           return_attachments - If true, attachments are added to the replica and returned in
            the response. Otherwise, attachments are not included. The default is false. This
            parameter is only applicable if the feature service has attachments.
           return_attachments_databy_url -  If true, a reference to a URL will be provided for
            each attachment returned from create method. Otherwise, attachments are embedded
            in the response. The default is true. This parameter is only applicable if the
            feature service has attachments and if return_attachments is true.
           asynchronous - If true, the request is processed as an asynchronous job, and a URL is
            returned that a client can visit to check the status of the job. See the topic on
            asynchronous usage for more information. The default is false.
           attachments_sync_direction - Client can specify the attachmentsSyncDirection when
            creating a replica. AttachmentsSyncDirection is currently a createReplica property
            and cannot be overridden during sync.
            Values: none, upload, bidirectional
           sync_model - this parameter is used to indicate that the replica is being created for
            per-layer sync or per-replica sync. To determine which model types are supported by a
            service, query the supportsPerReplicaSync, supportsPerLayerSync, and supportsSyncModelNone
            properties of the Feature Service. By default, a replica is created for per-replica sync.
            If syncModel is perReplica, the syncDirection specified during sync applies to all layers
            in the replica. If the syncModel is perLayer, the syncDirection is defined on a layer-by-layer
            basis.

            If syncModel is perReplica, the response will have replicaServerGen. A perReplica syncModel
            requires the replicaServerGen on sync. The replicaServerGen tells the server the point
            in time from which to send back changes. If syncModel is perLayer, the response will include
            an array of server generation numbers for the layers in layerServerGens. A perLayer sync
            model requires the layerServerGens on sync. The layerServerGens tell the server the point
            in time from which to send back changes for a specific layer. sync_model=none can be used
            to export the data without creating a replica. Query the supportsSyncModelNone property
            of the feature service to see if this model type is supported.

            See the RollbackOnFailure and Sync Models topic for more details.
            Values: perReplica | perLayer | none
            Example: syncModel=perLayer
           data_format - The format of the replica geodatabase returned in the response. The
            default is json.
            Values: filegdb, json, sqlite, shapefile
           replica_options - This parameter instructs the create operation to create a
            new replica based on an existing replica definition (refReplicaId). It can be used
            to specify parameters for registration of existing data for sync. The operation
            will create a replica but will not return data. The responseType returned in the
            create response will be esriReplicaResponseTypeInfo.
           wait - if async, wait to pause the process until the async operation is completed.
           out_path - folder path to save the file
        """
        return self._fs._create_replica(replica_name=replica_name,
                                        layers=layers,
                                        layer_queries=layer_queries,
                                        geometry_filter=geometry_filter,
                                        replica_sr=replica_sr,
                                        transport_type=transport_type,
                                        return_attachments=return_attachments,
                                        return_attachments_data_by_url=return_attachments_databy_url,
                                        asynchronous=asynchronous,
                                        sync_direction=sync_direction, # MISSING PARAMETER
                                        target_type=target_type, # Missing Parameter
                                        attachments_sync_direction=attachments_sync_direction,
                                        sync_model=sync_model,
                                        data_format=data_format,
                                        replica_options=replica_options,
                                        wait=wait,
                                        out_path=out_path)

    # ----------------------------------------------------------------------
    def synchronize(self,
                    replica_id,
                    transport_type="esriTransportTypeUrl",
                    replica_server_gen=None,
                    return_ids_for_adds=False,
                    edits=None,
                    return_attachment_databy_url=False,
                    asynchronous=False,
                    sync_direction="snapshot",
                    sync_layers="perReplica",
                    edits_upload_id=None,
                    edits_upload_format=None,
                    data_format="json",
                    rollback_on_failure=True):
        """
        synchronizes replica with feature layer collection
        http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000000vv000000
        """
        # TODO:
        return self._fs._synchronize_replica(replica_id=replica_id,
                                             transport_type=transport_type,
                                             replica_server_gen=replica_server_gen,
                                             return_ids_for_adds=return_ids_for_adds,
                                             edits=edits,
                                             return_attachment_databy_url=return_attachment_databy_url,
                                             asynchronous=asynchronous,
                                             sync_direction=sync_direction,
                                             sync_layers=sync_layers,
                                             edits_upload_id=edits_upload_id,
                                             edits_upload_format=edits_upload_format,
                                             data_format=data_format,
                                             rollback_on_failure=rollback_on_failure,
                                             close_replica=False,
                                             out_path=None)
    def create_replica_item(self,
                            replica_name,
                            item,
                            destination_gis,
                            layers=None,
                            extent=None):
        """creates a replicated service from a parent to another GIS"""
        import tempfile
        import os
        from ..gis import Item
        fs = item.layers[0].container
        if layers is None:
            ls = fs.properties['layers']
            ts = fs.properties['tables']
            layers = ""
            for i in ls + ts:
                layers += str(i['id'])
        if extent is None:
            extent = fs.properties['fullExtent']
            if 'spatialReference' in extent:
                del extent['spatialReference']
        out_path = tempfile.gettempdir()
        from . import FeatureLayerCollection
        isinstance(fs, FeatureLayerCollection)
        db = fs._create_replica(replica_name=replica_name,
                            layers=layers,
                            geometry_filter=extent,
                            attachments_sync_direction=None,
                            transport_type="esriTransportTypeUrl",
                            return_attachments=True,
                            return_attachments_data_by_url=True,
                            asynchronous=True,
                            sync_model="perLayer",
                            target_type="server",
                            data_format="sqlite",
                            #target_type="server",
                            out_path=out_path,
                            wait=True)
        if os.path.isfile(db) == False:
            raise Exception("Could not create the replica")
        destination_content = destination_gis.content
        item = destination_content.add(item_properties= {"type" : "SQLite Geodatabase",
                                                    "tags" : "replication",
                                                    "title" : replica_name},
                                  data=db)
        published = item.publish()
        return published

    def sync_replicated_items(self, parent, child, replica_name):
        """
        synchronizes two replicated items between portals

        Paramters:
         :parent: arcgis.gis.Item class that points to a feature service
          who is the parent (source) dataset.
         :child: arcgis.gis.Item class that points to the child replica
         :replica_name: name of the replica to synchronize
        Output:
         boolean value. True means service is up to date/synchronized,
         False means the synchronization failed.


        """
        from ..gis import Item
        if isinstance(parent, Item) == False:
            raise ValueError("parent must be an Item")
        if isinstance(child, Item) == False:
            raise ValueError("child must be an Item")
        child_fs = child.layers[0].container
        parent_fs = parent.layers[0].container
        child_replicas = child_fs.replicas
        parent_replicas = parent_fs.replicas
        if child_replicas and \
           parent_replicas:
            child_replica_id = None
            parent_replica_id = None
            child_replica = None
            parent_replica = None
            for replica in child_replicas.get_list():
                if replica['replicaName'].lower() == replica_name.lower():
                    child_replica_id = replica['replicaID']
                    break
            for replica in parent_replicas.get_list():
                if replica['replicaName'].lower() == replica_name.lower():
                    parent_replica_id = replica['replicaID']
                    break
            if child_replica_id and \
               parent_replica_id:
                import tempfile, os
                child_replica = child_replicas.get(replica_id=child_replica_id)
                parent_replica = parent_replicas.get(replica_id=parent_replica_id)
                delta = parent_fs._synchronize_replica(replica_id=parent_replica_id,
                                                       transport_type="esriTransportTypeUrl",
                                                       close_replica=False,
                                                       return_ids_for_adds=False,
                                                       return_attachment_databy_url=True,
                                                       asynchronous=False,
                                                       sync_direction="download",
                                                       sync_layers=parent_replica['layerServerGens'],
                                                       edits_upload_format="sqlite",
                                                       data_format="sqlite",
                                                       rollback_on_failure=False,
                                                       out_path=tempfile.gettempdir())
                if os.path.isfile(delta) == False:
                    return True
                work, message = child_fs.upload(path=delta)
                if isinstance(message, dict) and \
                   'item' in message and \
                   'itemID' in message['item']:
                    syncLayers_child = child_replica['layerServerGens']
                    syncLayers_parent = parent_replica['layerServerGens']
                    for i in range(len(syncLayers_parent)):
                        syncLayers_child[i]['serverSibGen'] = syncLayers_parent[i]['serverGen']
                    child_fs._synchronize_replica(
                        replica_id=child_replica_id,
                        sync_layers=syncLayers_child,
                        sync_direction=None,
                        edits_upload_id=message['item']['itemID'],
                        return_ids_for_adds=False,
                        data_format="sqlite",
                        asynchronous=False,
                        edits_upload_format="sqlite")
                    return True
                else:
                    return False
            else:
                raise ValueError("Could not find replica name %s in both services" % replica_name)
        else:
            return False
        return False

class FeatureLayerCollectionManager(_GISResource):
    """
    Allows updating the definition (if access permits) of a feature layer collection.
    This class is not created by users directly.
    An instance of this class, called 'manager', is available as a property of the FeatureLayerCollection object.

    Users call methods on this 'manager' object to manage the feature layer collection.
    """

    def __init__(self, url, gis=None, fs=None):
        super(FeatureLayerCollectionManager, self).__init__(url, gis)
        self._fs = fs
        self._populate_layers()

    def _populate_layers(self):
        """
        populates layers and tables in the managed feature service
        """
        layers = []
        tables = []

        try:
            for layer in self.properties.layers:
                layers.append(FeatureLayerManager(self.url + '/' + str(layer['id']), self._gis))
        except:
            pass

        try:
            for table in self.properties.tables:
                tables.append(FeatureLayerManager(self.url + '/' + str(table['id']), self._gis))
        except:
            pass

        self.layers = layers
        self.tables = tables
    # ----------------------------------------------------------------------
    def refresh(self):
        """ refreshes a feature layer collection """
        params = {"f": "json"}
        refresh_url = self._url + "/refresh"
        res = self._con.post(refresh_url, params)

        super(FeatureLayerCollectionManager, self)._refresh()
        self._populate_layers()

        self._fs._refresh()
        self._fs._populate_layers()

        return res
    #----------------------------------------------------------------------
    def _generate_mapservice_definition(self):
        """
        This operation returns a map service JSON that can be used to
        create a service.

        If a service does not support this operation, None is returned.

        :returns:
           dictionary
        """
        params = {
            'f' : 'json',
        }
        url = "%s/generateMapServiceDefinition" % self._url
        try:
            res = self._con.post(url, params)
        except:
            res = None
        return res
    # ----------------------------------------------------------------------
    def create_view(self,
                    name,
                    spatial_reference=None,
                    extent=None,
                    allow_schema_changes=True,
                    updateable=True,
                    capabilities="Query",
                    view_layers=None):
        """
        Creates a view of an existing feature service. You can create a view, if you need a different view of the data
        represented by a hosted feature layer, for example, you want to apply different editor settings, apply different
        styles or filters, define which features or fields are available, or share the data to different groups than
        the hosted feature layer  create a hosted feature layer view of that hosted feature layer.

        When you create a feature layer view, a new hosted feature layer item is added to Content. This new layer is a
        view of the data in the hosted feature layer, which means updates made to the data appear in the hosted feature
        layer and all of its hosted feature layer views. However, since the view is a separate layer, you can change
        properties and settings on this item separately from the hosted feature layer from which it is created.

        For example, you can allow members of your organization to edit the hosted feature layer but share a read-only
        feature layer view with the public.

        To learn more about views visit: https://doc.arcgis.com/en/arcgis-online/share-maps/create-hosted-views.htm

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        name                     Required string. Name of the new view item
        --------------------     --------------------------------------------------------------------
        spatial_reference        Optional dict. Specify the spatial reference of the view
        --------------------     --------------------------------------------------------------------
        extent                   Optional dict. Specify the extent of the view
        --------------------     --------------------------------------------------------------------
        allow_schema_changes     Optional bool. Default is True. Determines if a view can alter a
                                 service's schema.
        --------------------     --------------------------------------------------------------------
        updateable               Optional bool. Default is True. Determines if view can update values
        --------------------     --------------------------------------------------------------------
        capabilities             Optional string. Specify capabilities as a comma separated string.
                                 For example "Query, Update, Delete". Default is 'Query'.
        --------------------     --------------------------------------------------------------------
        view_layers              Optional list. Specify list of layers present in the FeatureLayerCollection
                                 that you want in the view.
        ====================     ====================================================================

        .. code-block:: python  (optional)

           USAGE EXAMPLE: Create a veiw from a hosted feature layer

           crime_fl_item = gis.content.search("2012 crime")[0]
           crime_flc = FeatureLayerCollection.fromitem(crime_fl_item)

           # Create a view with just the first layer
           crime_view = crime_flc.manager.create_view(name='Crime in 2012", updateable=False,
                                                        view_layers=[crime_flc.layers[0]])

        .. code-block:: python (optional)

            USAGE EXAMPLE: Create an editable view

            crime_fl_item = gis.content.search("2012 crime")[0]
            crime_flc = FeatureLayerCollection.fromitem(crime_fl_item)
            crime_view = crime_flc.manager.create_view(name=uuid.uuid4().hex[:9], # create random name
                                                       updateable=True,
                                                       allow_schema_changes=False,
                                                       capabilities="Query,Update,Delete")

        :return:
            Returns the newly created item for the view.
        """

        import os
        from . import FeatureLayerCollection
        gis = self._gis
        content = gis.content
        if 'serviceItemId' not in self.properties:
            raise Exception("A registered hosted feature service is required to use create_view")
        item_id = self.properties['serviceItemId']
        item = content.get(itemid=item_id)
        url = item.url
        fs = FeatureLayerCollection(url=url, gis=gis)
        if gis._url.lower().find("sharing/rest") < 0:
            url = gis._url + "/sharing/rest"
        else:
            url = gis._url

        if 'id' in gis.properties.user:
            me = gis.users.me.username
            #me = gis.properties.user.id
        else:
            me = gis.users.me.username
        url = "%s/content/users/%s/createService" % (url, me)
        params = {
            "f" : "json",
            "isView" : True,
            "createParameters" :json.dumps({"name": name,
                                            "isView":True,
                                            "sourceSchemaChangesAllowed": allow_schema_changes,
                                            "isUpdatableView": updateable,
                                            "spatialReference":spatial_reference or fs.properties['spatialReference'],
                                            "initialExtent": extent or fs.properties['initialExtent'],
                                            "capabilities":capabilities or fs.properties['capabilties']}),
            "outputType" : "featureService"
        }
        res = gis._con.post(path=url, postdata=params)
        view = content.get(res['itemId'])
        fs_view = FeatureLayerCollection(url=view.url, gis=gis)
        add_def = {
            "layers" : []
        }
        if view_layers is None:
            for lyr in fs.layers:
                add_def['layers'].append(
                    {
                        "adminLayerInfo" : {
                            "viewLayerDefinition" :
                        {
                            "sourceServiceName" : os.path.basename(os.path.dirname(fs.url)),
                            "sourceLayerId" : lyr.manager.properties['id'],
                            "sourceLayerFields" :  "*"
                        }
                        },
                        "name" : lyr.manager.properties['name']
                    })
        elif isinstance(view_layers, list):
            for lyr in view_layers:
                add_def['layers'].append(
                    {
                        "adminLayerInfo": {"viewLayerDefinition":
                                               {"sourceServiceName": os.path.basename(os.path.dirname(fs.url)),
                                                "sourceLayerId": lyr.manager.properties['id'],
                                                "sourceLayerFields": "*"}
                                           },
                        "name": lyr.manager.properties['name']
                    })
        else:
            add_def = view_layers
        fs_view.manager.add_to_definition(add_def)
        return content.get(res['itemId'])
    # ----------------------------------------------------------------------
    def add_to_definition(self, json_dict):
        """
           The add_to_definition operation supports adding a definition
           property to a hosted feature layer collection service. The result of this
           operation is a response indicating success or failure with error
           code and description.

           This function will allow users to change or add additional values
           to an already published service.

           Input:
              json_dict - part to add to host service.  The part format can
                          be derived from the properties property.  For
                          layer level modifications, run updates on each
                          individual feature service layer object.
           Output:
              JSON message as dictionary
        """
        if isinstance(json_dict, PropertyMap):
            json_dict = dict(json_dict)

        params = {
            "f": "json",
            "addToDefinition": json.dumps(json_dict),
            "async": json.dumps(False)
        }
        adddefn_url = self._url + "/addToDefinition"
        old_ref = self._con._referer
        self._con._referer = "http"
        res = self._con.post(adddefn_url, params)
        self.refresh()
        self._con._referer = old_ref
        return res

    # ----------------------------------------------------------------------
    def update_definition(self, json_dict):
        """
           The update_definition operation supports updating a definition
           property in a hosted feature layer collection service. The result of this
           operation is a response indicating success or failure with error
           code and description.

           Input:
              json_dict - part to add to host service.  The part format can
                          be derived from the properties property.  For
                          layer level modifications, run updates on each
                          individual feature service layer object.
           Output:
              JSON Message as dictionary
        """
        definition = None
        if json_dict is not None:

            if isinstance(json_dict, PropertyMap):
                definition = dict(json_dict)
            if isinstance(json_dict, collections.OrderedDict):
                definition = json_dict
            else:

                definition = collections.OrderedDict()
                if 'hasStaticData' in json_dict:
                    definition['hasStaticData'] = json_dict['hasStaticData']
                if 'allowGeometryUpdates' in json_dict:
                    definition['allowGeometryUpdates'] = json_dict['allowGeometryUpdates']
                if 'capabilities' in json_dict:
                    definition['capabilities'] = json_dict['capabilities']
                if 'editorTrackingInfo' in json_dict:
                    definition['editorTrackingInfo'] = collections.OrderedDict()
                    if 'enableEditorTracking' in json_dict['editorTrackingInfo']:
                        definition['editorTrackingInfo']['enableEditorTracking'] = json_dict['editorTrackingInfo'][
                            'enableEditorTracking']

                    if 'enableOwnershipAccessControl' in json_dict['editorTrackingInfo']:
                        definition['editorTrackingInfo']['enableOwnershipAccessControl'] = \
                            json_dict['editorTrackingInfo']['enableOwnershipAccessControl']

                    if 'allowOthersToUpdate' in json_dict['editorTrackingInfo']:
                        definition['editorTrackingInfo']['allowOthersToUpdate'] = json_dict['editorTrackingInfo'][
                            'allowOthersToUpdate']

                    if 'allowOthersToDelete' in json_dict['editorTrackingInfo']:
                        definition['editorTrackingInfo']['allowOthersToDelete'] = json_dict['editorTrackingInfo'][
                            'allowOthersToDelete']

                    if 'allowOthersToQuery' in json_dict['editorTrackingInfo']:
                        definition['editorTrackingInfo']['allowOthersToQuery'] = json_dict['editorTrackingInfo'][
                            'allowOthersToQuery']
                    if isinstance(json_dict['editorTrackingInfo'], dict):
                        for key, val in json_dict['editorTrackingInfo'].items():
                            if key not in definition['editorTrackingInfo']:
                                definition['editorTrackingInfo'][key] = val
                if isinstance(json_dict, dict):
                    for key, val in json_dict.items():
                        if key not in definition:
                            definition[key] = val

        params = {
            "f": "json",
            "updateDefinition": json.dumps(obj=definition, separators=(',', ':')),
            "async": False
        }
        u_url = self._url + "/updateDefinition"
        old_ref = self._con._referer
        self._con._referer = "http"
        res = self._con.post(u_url, params)
        self._con._referer = old_ref
        self.refresh()
        return res

    # ----------------------------------------------------------------------
    def delete_from_definition(self, json_dict):
        """
        The delete_from_definition operation supports deleting a
        definition property from a hosted feature layer collection service. The result of
        this operation is a response indicating success or failure with
        error code and description.
        See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Delete_From_Definition_Feature_Service/02r30000021w000000/ # noqa
        for additional information on this function.
        Input:
          json_dict - part to add to host service.  The part format can
                      be derived from the properties property.  For
                      layer level modifications, run updates on each
                      individual feature service layer object.  Only
                      include the items you want to remove from the
                      FeatureService or layer.

        Output:
          JSON Message as dictionary

        """
        params = {
            "f": "json",
            "deleteFromDefinition": json.dumps(json_dict),
            "async": False
        }
        u_url = self._url + "/deleteFromDefinition"

        res = self._con.post(u_url, params)
        self.refresh()
        return res

    # ----------------------------------------------------------------------
    def overwrite(self, data_file):
        """
        Overwrite all the features and layers in a hosted feature layer collection service. This operation removes
        all features but retains the properties (such as metadata, itemID) and capabilities configured on the service.
        There are some limits to using this operation:
            1. Only hosted feature layer collection services can be overwritten
            2. The original data used to publish this layer should be available on the portal
            3. The data file used to overwrite should be of the same format and filename as the original that was used to
            publish the layer
            4. The schema (column names, column data types) of the data_file should be the same as original. You can have
            additional or fewer rows (features).

        In addition to overwriting the features, this operation also updates the data of the item used to published this
        layer.

        :param data: path to data_file used to overwrite the hosted feature layer collection
        :return: JSON message as dictionary such as {'success':True} or {'error':'error message'}
        """

        # region Get Item associated with the service
        if 'serviceItemId' in self.properties.keys():
            feature_layer_item = self._gis.content.get(self.properties['serviceItemId'])
        else:
            return {'error': 'Can only overwrite a hosted feature layer collection'}
        # endregion

        # region find data item related to this hosted feature layer
        related_data_items = feature_layer_item.related_items('Service2Data', 'forward')
        if len(related_data_items) > 0:
            related_data_item = related_data_items[0]
        else:
            return {'error': 'Cannot find related data item used to publish this feature layer'}

        # endregion
        params = None
        if related_data_item.type in ['CSV', 'Shapefile', 'File Geodatabase'] and \
           self._gis._portal.is_arcgisonline == False:
            params = {
                "name" : related_data_item.name,
                "title" : related_data_item.title,
                "tags" : related_data_item.tags,
                "type" : related_data_item.type,
                "overwrite" : True,
                "overwriteService" : "on",
                "useDescription" : "on"
            }
            lyr_url_info = "%s/layers" % feature_layer_item.layers[0].container._url
            fs_url = "%s" % feature_layer_item.layers[0].container._url
            layer_info = self._gis._con.get(lyr_url_info, {'f' : 'json'})
            [lyr.pop('fields') for lyr in layer_info['layers']]
            [lyr.pop('fields') for lyr in layer_info['tables']]
            feature_service_def = self._gis._con.get(fs_url, {'f' : 'json'})
            feature_service_def['tables'] = []
            feature_service_def['layers'] = []
            feature_service_def.update(layer_info)
            publish_parameters = feature_service_def
            publish_parameters['name'] = feature_layer_item.title
            publish_parameters['_ssl'] = False
            for idx, lyr in enumerate(publish_parameters['layers']):
                lyr['parentLayerId'] = -1
                for k in {'sourceSpatialReference', 'isCoGoEnabled',
                          'parentLayer', 'isDataArchived', 'cimVersion'}:
                    lyr.pop(k, None)
            for idx, lyr in enumerate(publish_parameters['tables']):
                lyr['parentLayerId'] = -1
                for k in {'sourceSpatialReference', 'isCoGoEnabled',
                          'parentLayer', 'isDataArchived', 'cimVersion'}:
                    lyr.pop(k, None)


        # region construct publishParameters dictionary
        elif related_data_item.type in ['CSV', 'Shapefile', 'File Geodatabase'] and \
           self._gis._portal.is_arcgisonline:
            # construct a full publishParameters that is a combination of existing Feature Layer definition
            # and original publishParameters.json used for publishing the service the first time

            # get old publishParameters.json
            path = "content/items/" + feature_layer_item.itemid + "/info/publishParameters.json"
            postdata = {'f': 'json'}

            old_publish_parameters = self._gis._con.post(path, postdata)

            # get FeatureServer definition
            feature_service_def = dict(self.properties)

            # Get definition of each layer and table, remove fields in the dict
            layers_dict = []
            tables_dict = []
            for layer in self.layers:
                layer_def = dict(layer.properties)
                if 'fields' in layer_def.keys():
                    dump = layer_def.pop("fields")
                layers_dict.append(layer_def)

            for table in self.tables:
                table_def = dict(table.properties)
                if 'fields' in table_def.keys():
                    dump = table_def.pop('fields')
                tables_dict.append(table_def)

            # Splice the detailed table and layer def with FeatuerServer def
            feature_service_def['layers'] = layers_dict
            feature_service_def['tables'] = tables_dict
            from pathlib import Path
            service_name = Path(self.url).parts[-2]  # get service name from url
            feature_service_def['name'] = service_name

            # combine both old publish params and full feature service definition
            publish_parameters = feature_service_def
            publish_parameters.update(old_publish_parameters)
        else:
            # overwriting a SD case - no need for detailed publish parameters
            publish_parameters = None

        # endregion

        #region Perform overwriting
        if related_data_item.update(item_properties=params, data=data_file):
            published_item = related_data_item.publish(publish_parameters, overwrite=True)
            if published_item is not None:
                return {'success': True}
            else:
                return {'error': 'Unable to overwrite the hosted feature layer collection'}
        else:
            return {'error': 'Unable to update related data item with new data'}

    # ----------------------------------------------------------------------

    def _gen_overwrite_publishParameters(self, flc_item):
        """
        This internal method generates publishParameters for overwriting a hosted feature layer collection. This is used
        by Item.publish() method when user wants to originate the overwrite process from the data item instead of
        the hosted feature layer.

        :param flc_item: The Feature Layer Collection Item object that is being overwritten
        :return: JSON message as dictionary with to be used as publishParameters payload in the publish REST call.
        """

        # region construct publishParameters dictionary
        # construct a full publishParameters that is a combination of existing Feature Layer definition
        # and original publishParameters.json used for publishing the service the first time

        # get old publishParameters.json
        path = "content/items/" + flc_item.itemid + "/info/publishParameters.json"
        postdata = {'f': 'json'}

        old_publish_parameters = self._gis._con.post(path, postdata)

        # get FeatureServer definition
        feature_service_def = dict(self.properties)

        # Get definition of each layer and table, remove fields in the dict
        layers_dict = []
        tables_dict = []
        for layer in self.layers:
            layer_def = dict(layer.properties)
            if 'fields' in layer_def.keys():
                dump = layer_def.pop("fields")
            layers_dict.append(layer_def)

        for table in self.tables:
            table_def = dict(table.properties)
            if 'fields' in table_def.keys():
                dump = table_def.pop('fields')
            tables_dict.append(table_def)

        # Splice the detailed table and layer def with FeatuerServer def
        feature_service_def['layers'] = layers_dict
        feature_service_def['tables'] = tables_dict
        from pathlib import Path
        service_name = Path(self.url).parts[-2]  # get service name from url
        feature_service_def['name'] = service_name

        # combine both old publish params and full feature service definition
        publish_parameters = feature_service_def
        publish_parameters.update(old_publish_parameters)

        # endregion

        return publish_parameters

class FeatureLayerManager(_GISResource):
    """
    Allows updating the definition (if access permits) of a feature layer. This class is not created by users
    directly.
    An instance of this class, called 'manager', is available as a property of the FeatureLayer object,
    if the layer can be managed by the user.
    Users call methods on this 'manager' object to manage the feature layer.
    """

    def __init__(self, url, gis=None):
        super(FeatureLayerManager, self).__init__(url, gis)
        self._hydrate()

    # ----------------------------------------------------------------------
    @classmethod
    def fromitem(cls, item, layer_id=0):
        """
        Creates a FeatureLayerManager object from a GIS Item.
        The type of item should be a 'Feature Service' that represents a FeatureLayerCollection.
        The layer_id is the id of the layer in feature layer collection (feature service).
        """
        if item.type != "Feature Service":
            raise TypeError("item must be a of type Feature Service, not " + item.type)
        from arcgis.features import FeatureLayer
        return FeatureLayer.fromitem(item, layer_id).manager

    # ----------------------------------------------------------------------
    def refresh(self):
        """ refreshes a service """
        params = {"f": "json"}
        u_url = self._url + "/refresh"
        res = self._con.post(u_url, params)

        super(FeatureLayerManager, self)._refresh()

        return res

    # ----------------------------------------------------------------------
    def add_to_definition(self, json_dict):
        """
           The addToDefinition operation supports adding a definition
           property to a hosted feature layer. The result of this
           operation is a response indicating success or failure with error
           code and description.

           This function will allow users to change add additional values
           to an already published service.

           Input:
              json_dict - part to add to host service.  The part format can
                          be derived from the asDictionary property.  For
                          layer level modifications, run updates on each
                          individual feature service layer object.
           Output:
              JSON message as dictionary
        """

        if isinstance(json_dict, PropertyMap):
            json_dict = dict(json_dict)

        params = {
            "f": "json",
            "addToDefinition": json.dumps(json_dict),
            # "async" : False
        }
        u_url = self._url + "/addToDefinition"

        res = self._con.post(u_url, params)
        self.refresh()
        return res

    # ----------------------------------------------------------------------
    def update_definition(self, json_dict):
        """
           The updateDefinition operation supports updating a definition
           property in a hosted feature layer. The result of this
           operation is a response indicating success or failure with error
           code and description.

           Input:
              json_dict - part to add to host service.  The part format can
                          be derived from the asDictionary property.  For
                          layer level modifications, run updates on each
                          individual feature service layer object.
           Output:
              JSON Message as dictionary
        """

        if isinstance(json_dict, PropertyMap):
            json_dict = dict(json_dict)

        params = {
            "f": "json",
            "updateDefinition": json.dumps(json_dict),
            "async": False
        }

        u_url = self._url + "/updateDefinition"

        res = self._con.post(u_url, params)
        self.refresh()
        return res

    # ----------------------------------------------------------------------
    def delete_from_definition(self, json_dict):
        """
           The deleteFromDefinition operation supports deleting a
           definition property from a hosted feature layer. The result of
           this operation is a response indicating success or failure with
           error code and description.
           See: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Delete_From_Definition_Feature_Service/02r30000021w000000/ # noqa
           for additional information on this function.
           Input:
              json_dict - part to add to host service.  The part format can
                          be derived from the asDictionary property.  For
                          layer level modifications, run updates on each
                          individual feature service layer object.  Only
                          include the items you want to remove from the
                          FeatureService or layer.

           Output:
              JSON Message as dictionary

        """

        if isinstance(json_dict, PropertyMap):
            json_dict = dict(json_dict)

        params = {
            "f": "json",
            "deleteFromDefinition": json.dumps(json_dict)
        }
        u_url = self._url + "/deleteFromDefinition"

        res = self._con.post(u_url, params)
        self.refresh()
        return res

    # ----------------------------------------------------------------------
    def truncate(self,
                 attachment_only=False,
                 asynchronous=False,
                 wait=True):
        """
           The truncate operation supports deleting all features or attachments
           in a hosted feature service layer. The result of this operation is a
           response indicating success or failure with error code and description.
           See: http://resources.arcgis.com/en/help/arcgis-rest-api/#/Truncate_Feature_Layer/02r3000002v0000000/ # noqa
           for additional information on this function.
           Input:
              attachment_only - Deletes all the attachments for this layer.
                                None of the layer features will be deleted
                                when attachmentOnly=true.
              asynchronous - Supports options for asynchronous processing. The
                      default format is false. It is recommended to set
                      async=true for larger datasets.
              wait - if async, wait to pause the process until the async
                     operation is completed.

           Output:
              JSON Message as dictionary

        """
        params = {
            "f": "json",
            "attachmentOnly": attachment_only,
            "async": asynchronous
        }
        u_url = self._url + "/truncate"

        if asynchronous:
            if wait:
                job = self._con.post(u_url, params)
                status = self._get_status(url=job['statusURL'])
                while status['status'] not in ("Completed", "CompletedWithErrors", "Failed"):
                    # wait before checking again
                    time.sleep(2)
                    status = self._get_status(url=job['statusURL'])

                res = status
                self.refresh()
            else:
                res = self._con.post(u_url, params)
                # Leave calling refresh to user since wait is false
        else:
            res = self._con.post(u_url, params)
            self.refresh()
        return res

    # ----------------------------------------------------------------------
    def _get_status(self, url):
        """gets the status when exported async set to True"""
        params = {"f": "json"}
        url += "/status"
        return self._con.get(url, params)



"""
The arcgis.features module contains types and functions for working with features and feature layers in the GIS.

Entities located in space with a geometrical representation (such as points, lines or polygons) and a set of properties
can be represented as features. The arcgis.features module is used for working with feature data, feature layers and
collections of feature layers in the GIS. It also contains the spatial analysis functions which operate against
feature data.

In the GIS, entities located in space with a set of properties can be represented as features.  Features are stored as
feature classes, which represent a set of features located using a single spatial type (point, line, polygon) and a
common set of properties.  This is the geographic extension of the classic tabular or relational representation for
entities - a set of entities is modelled as rows in a table.  Tables represent entity classes with uniform
properties.  In addition to working with entities with location as features, the system can also work with
non-spatial entities as rows in tables.  The system can also model relationships between entities using properties
which act as primary and foreign keys.  A collection of feature classes and tables, with the associated
relationships among the entities, is a feature layer collection. FeatureLayerCollections are one of the dataset types
contained in a Datastore.  Finally, features are not simply entities in a dataset.  Features have a visual
representation and user experience - on a map, in a 3D scene, as entities with a property sheet or popups.
"""

from .feature import Feature, FeatureSet, FeatureCollection
from .layer import FeatureLayer, Table, FeatureLayerCollection
from ._data.geodataset import SpatialDataFrame
from ._parcel import ParcelFabricManager
from ._utility import UtilityNetworkManager

from . import analyze_patterns
from . import enrich_data
from . import find_locations
from . import manage_data
from . import summarize_data
from . import use_proximity
from . import analysis


__all__ = ['Feature', 'FeatureSet',
           'FeatureCollection', 'FeatureLayer',
           'Table', 'FeatureLayerCollection']
try:
    from .geo import GeoAccessor, GeoSeriesAccessor
    __all__.extend(['GeoAccessor', 'GeoSeriesAccessor'])
except ImportError:
    pass





from arcgis.gis import GIS
from arcgis import env
from arcgis._impl.common._mixins import PropertyMap

########################################################################
class UtilityNetworkManager(object):
    """
    The Utility Network Service exposes analytic capabilities (tracing)
    as well as validation of network topology and management of
    subnetworks (managing sources, updating subnetworks, exporting
    subnetworks, and so on). The Utility Network Service is conceptually
    similar to the Network Analysis Service for transportation networks.

    =====================   ===========================================
    **Inputs**              **Description**
    ---------------------   -------------------------------------------
    url                     Required String. The web endpoint to the utility service.
    ---------------------   -------------------------------------------
    version                 Required Version. The `Version` class where the branch version will take place.
    ---------------------   -------------------------------------------
    gis                     Optional GIS. The `GIS` connection object.
    =====================   ===========================================


    """
    _con = None
    _gis = None
    _url = None
    _version = None
    _property = None
    _version_guid = None
    _version_name = None
    #----------------------------------------------------------------------
    def __init__(self, url, version, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        self._gis = gis
        self._con = gis._portal.con
        self._url =  url
        self._version = version
        self._version_guid = version._guid
        self._version_name = version.properties.versionName
    #----------------------------------------------------------------------
    def _init(self):
        """initializer"""
        try:
            res = self._con.get(self._url, {'f':'json'})
            self._property = PropertyMap(res)
        except Exception as e:
            self._property = PropertyMap({})

    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties for the service"""
        if self._property is None:
            self._init()
        return self._property
    #----------------------------------------------------------------------
    def trace(self,
              locations,
              trace_type,
              fields=None,
              moment=None,
              configuration=None,
              result_type=None
              ):
        """
        A trace refers to a pre-configured algorithm that systematically
        travels a network to return results. Generalized traces allow you to
        trace across multiple types of domain networks. For example, running
        a Connected trace from your electric network through to your gas
        network. An assortment of options is provided with trace to support
        various analytic work flows. All traces use the network topology to
        read cached information about network features. This can improve
        performance of complex traces on large networks. Trace results are
        not guaranteed to accurately represent a utility network when dirty
        areas are present. The network topology must be validated to ensure
        it reflects the most recent edits or updates made to the network.


        """
        url = "%s/trace" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            "traceType" : trace_type,
            "moment" : moment,
            "traceLocations" : locations,
            "traceConfiguration" : configuration,
            "resultFields" : fields,
            "resultType" : result_type
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def disable_subnetwork_controller(self,
                                      network_source_id,
                                      global_id,
                                      terminal_id):
        """
        A subnetwork controller (or simply, a source or a sink) is the
        origin (or destination) of resource flow for a subpart of the
        network. Examples of subnetwork controllers are circuit breakers in
        electric networks, or town border stations in gas networks.
        Subnetwork controllers correspond to devices that have the
        Subnetwork Controller network capability set. A source is removed
        with `disable_subnetwork_controller`.


        """

        url = "%s/disableSubnetworkController" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            'networkSourceId' : network_source_id,
            'featureGlobalId' : global_id,
            'terminalId' : terminal_id
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def enable_subnetwork_controller(self,
                                     network_source_id,
                                     global_id,
                                     terminal_id,
                                     subnetwork_controller_name,
                                     tier_name,
                                     subnetwork_name=None,
                                     description=None,
                                     notes=None
                                     ):
        """
        A subnetwork controller is the origin (or destination) of resource
        flow for a subpart of the network (e.g., a circuit breaker in
        electric networks, or a town border station in gas networks).
        Controllers correspond to Devices that have the Subnetwork
        Controller network capability set.


        """

        url = "%s/enableSubnetworkController" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            'networkSourceId' : network_source_id,
            'featureGlobalId' : global_id,
            'terminalID' : terminal_id,
            'subnetworkControllerName' : subnetwork_controller_name,
            'subnetworkName' : subnetwork_name,
            'tierName' : tier_name,
            'description' : description,
            'notes' : notes
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def export_subnetwork(self,
                          domain_name,
                          tier_name,
                          subnetwork_name,
                          trace_configuration=None,
                          export_acknowlegement=False,
                          fields=None,
                          result_type=None,
                          moment=None):
        """
        The `export_subnetwork` operation is used to export information
        about a subnetwork into a JSON file. That information can then be
        consumed by outside systems such as outage management and asset
        tracking. The exportSubnetwork operation allows you to delete
        corresponding rows in the Subnetwork Sources table as long as the
        IsDeleted attribute is set to True. This indicates a source feeding
        the subnetwork has been removed.

        """

        url = "%s/exportSubnetwork" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            "moment" : moment,
            "domainNetworkName" : domain_name,
            "tierName" : tier_name,
            "subnetworkName" : subnetwork_name,
            "exportAcknowledgement" : export_acknowlegement,
            "traceConfiguration" : trace_configuration,
            "resultFields" : fields,
            "resultType" : result_type
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def query_network_moments(self,
                              moments_to_return="fullValidateTopology",
                              moment=None
                             ):
        """
        The `query_network_moments` operation returns the moments related
        to the network topology and operations against the topology. This
        includes when the topology was initially enabled, when it was last
        validated, when the topology was last disabled (and later enabled),
        and when the definition of the utility network was last modified.
        """
        url = "%s/queryNetworkMoments" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            "momentsToReturn" : moments_to_return,
            "moment" : moment
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def query_overrides(self,
                        attribute_ids=None,
                        all_attributes=False,
                        all_connectivity=False):
        """
        Network attributes support the ability to have their values
        overridden without having to edit features and validate the network
        topology (build the index). The utility network also supports the
        ability to place ephemeral connectivity (e.g., jumpers in an
        electrical network) between two devices or junctions without having
        to edit features or connectivity associations and validate the
        network topology (build the index). This operation allows the
        client to query all the overrides associated with the network
        attributes (by network attribute id). In addition, all connectivity
        overrides are returned.
        """
        url = "%s/queryOverrides" % self._url
        params = {
            "f" : "json",
            "attributeIDs" : attribute_ids,
            "allAttributes" : all_attributes,
            "allConnectivity" : all_connectivity
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def synthesize_association_geometries(self,
                                          attachment_associations=False,
                                          connectivity_associations=False,
                                          containment_associations=False,
                                          count=200,
                                          extent=False,
                                          out_sr=None,
                                          moment=None):
        """
        The `synthesize_association_geometries` operation is used to export
        geometries representing associations that are synthesized as line
        segments corresponding to the geometries of the devices at the
        endpoints. All features associated with an association must be in
        the specified extent in order for the geometry to be synthesized.
        If only zero or one of the devices/junctions intersects the extent,
        then no geometry will be synthesized.



        """
        url = "%s/synthesizeAssociationGeometries"
        params = {
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            "moment": moment,
            "attachmentAssociations": attachment_associations,
            "connectivityAssociations": connectivity_associations,
            "containmentAssociations": containment_associations,
            "maxGeometryCount":  count,
            "extent": extent,
            "outSR":  out_sr,
            "f" : "json"
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def update_is_connected(self):
        """

        Utility network features have an attribute called IsConnected that
        lets you know if a feature is connected to a source or sink, and
        therefore it could potentially be part of an existing subnetwork.
        The `update_is_connected` operation updates this attribute on
        features in the specified utility network. This operation can only
        be executed on the default version by the portal utility network
        owner.
        """
        url = "%s/updateIsConnected" % self._url
        params = {"f" : "json"}
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def update_subnetwork(self,
                          domain_name,
                          tier_name,
                          subnetwork_name=None,
                          all_subnetwork_tier=False,
                          continue_on_failure=False,
                          trace_configuration=None):
        """
        A subnetwork is updated by calling the `update_subnetwork` operation.
        With this operation, one or all of the subnetworks in a single tier
        can be updated. When a subnetwork is updated, four things can occur;
        the Subnetwork Name attribute is updated for all features in the
        subnetwork, the record representing the subnetwork inside the
        SubnetLine class is refreshed, the Subnetworks table is updated and
        finally diagrams are generated or updated for the subnetwork.

        :returns: Boolean

        """
        url = "%s/updateSubnetwork" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            "domainNetworkName" : domain_name,
            "tierName" : tier_name,
            "subnetworkName" : subnetwork_name,
            "allSubnetworksInTier" : all_subnetwork_tier,
            "continueOnFailure" : continue_on_failure,
            "traceConfiguration" : trace_configuration
        }
        return self._con.post(url, params)['success']
    #----------------------------------------------------------------------
    def validate_topology(self,
                          envelope,
                          run_async=False):
        """
        Validating the network topology for a utility network maintains
        consistency between feature editing space and network topology space.
        Validating a network topology may include all or a subset of the
        dirty areas present in the network. Validation of network topology
        is supported synchronously and asynchronously.
        """
        url = "%s/validateNetworkTopology" % self._url
        params = {
            "f" : "json",
            "gdbVersion" : self._version_name,
            "sessionId" : self._version_guid,
            "validateArea" : envelope,
            "async" : run_async
        }
        return self._con.post(url, params)['success']
    #----------------------------------------------------------------------
    def apply_overrides(self, adds=None, deletes=None):
        """
        Network attributes support the ability to have their values
        overridden without having to edit features and validate the network
        topology (build the index). The utility network also supports the
        ability to place ephemeral connectivity (for example, jumpers in an
        electrical network) between two devices or junctions without having
        to edit features or connectivity associations and validate the
        network topology (build the index). When specified by the client, a
        trace operation may optionally incorporate the network attribute
        and connectivity override values when the trace is run on.


        """
        url = "%s/applyOverrides"
        params = {'f' : 'json',
                  'adds' : adds,
                  'deletes' : deletes}
        return self._con.post(url, params)



import torch
import numpy as np
from torch import tensor
import torch
import math

def A(*a): return np.array(a[0]) if len(a)==1 else [np.array(o) for o in a]

imagenet_stats = A([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

mean = 255* np.array(imagenet_stats[0], dtype=np.float32)
std  = 255* np.array(imagenet_stats[1], dtype=np.float32)

norm = lambda x: (x-mean)/ std
denorm = lambda x: x * std + mean


def pred2dict(bb_np, score, cat_str, c):
    # convert to top left x,y bottom right x,y
    return {"x1": bb_np[1],
            "x2": bb_np[3],
            "y1": bb_np[0],
            "y2": bb_np[2],
            "score": score,
            "category": cat_str,
            "class": c}

def to_np(x):
    return x.cpu().numpy()

def load_weights(m, p):
    if p.split('.')[-1] == 'h5':
        sd = torch.load(p, map_location=lambda storage, loc: storage)
    elif p.split('.')[-1] == 'pth':
        # sd = torch.load(p)['model']
        sd = torch.load(p, map_location=lambda storage, loc: storage)['model']
    m.load_state_dict(sd)
    return m

def predict_(model, images, device):
    model = model.to(device)
    images = tensor(images).to(device)
    clas, bbox = model(images)
    return clas, bbox
    
def hw2corners(ctr, hw): 
        return torch.cat([ctr-hw/2, ctr+hw/2], dim=1)
    
def actn_to_bb(actn, anchors, grid_sizes):
        actn_bbs = torch.tanh(actn)
        actn_centers = (actn_bbs[:,:2]/2 * grid_sizes) + anchors[:,:2]
        actn_hw = (actn_bbs[:,2:]/2+1) * anchors[:,2:]
        return hw2corners(actn_centers, actn_hw)
    
def nms(boxes, scores, overlap=0.5, top_k=100):
    keep = scores.new(scores.size(0)).zero_().long()
    if boxes.numel() == 0: return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        keep[count] = i
        count += 1
        if idx.size(0) == 1: break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w*h
        # IoU = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        IoU = inter/union  # store result in iou
        # keep only elements with an IoU <= overlap
        idx = idx[IoU.le(overlap)]
    return keep, count

def get_nms_preds(b_clas, b_bb, idx, anchors, grid_sizes, classes, nms_overlap, thres):
    
    a_ic = actn_to_bb(b_bb[idx], anchors, grid_sizes)
    clas_pr, clas_ids = b_clas[idx].max(1)
    clas_pr = clas_pr.sigmoid()

    conf_scores = b_clas[idx].sigmoid().t().data

    out1, out2, cc = [], [], []
    for cl in range(1, len(conf_scores)):
        c_mask = conf_scores[cl] > thres
        if c_mask.sum() == 0: continue
        scores = conf_scores[cl][c_mask]
        l_mask = c_mask.unsqueeze(1).expand_as(a_ic)
        boxes = a_ic[l_mask].view(-1, 4)
        ids, count = nms(boxes.data, scores, nms_overlap, 50) # FIX- NMS overlap hardcoded
        ids = ids[:count]
        out1.append(scores[ids])
        out2.append(boxes.data[ids])
        cc.append([cl]*count)
    if cc == []:
        cc = [[0]]
    cc = tensor(np.concatenate(cc))
    if out1 == []:
        out1 = [torch.Tensor()]
    out1 = torch.cat(out1)
    if out2 == []:
        out2 = [torch.Tensor()]
    out2 = torch.cat(out2)
    bbox, clas, prs, thresh = out2, cc, out1, thres  # FIX- hardcoded threshold
    return predictions(bbox,
         to_np(clas), to_np(prs) if prs is not None else None, thresh, classes)

def predictions(bbox, clas=None, prs=None, thresh=0.3, classes=None): # FIX- take threshold from user
    #bb = [bb_hw(o) for o in bbox.reshape(-1,4)]
    if len(classes) is None:
        raise Exception("Classes are None")
    else:
        classes = ['bg'] + classes
    bb = bbox
    if prs is None:  prs  = [None]*len(bb)
    if clas is None: clas = [None]*len(bb)
    predictions = []
    for i, (b, c, pr) in enumerate(zip(bb, clas, prs)):
        if((b[2]>0) and (pr is None or pr > thresh)):
            cat_str = classes[c]
            score = pr * 100
            bb_np = to_np(b).astype('float64')
            predictions.append(pred2dict(bb_np, score, cat_str, c))
    return predictions

def detect_objects_image_space(model, tiles, anchors, grid_sizes, device, classes, nms_overlap, thres):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]
    img_normed = norm(tiles.transpose(0,2,3,1))


    clas, bbox = predict_(model, img_normed.transpose(0, 3, 1, 2), device)

    preds = { }

    for batch_idx in range(bbox.size()[0]):
        preds[batch_idx] = get_nms_preds(clas, bbox, batch_idx, anchors, grid_sizes, classes, nms_overlap, thres)

    batch_size = bbox.size()[0]
    side = math.sqrt(batch_size)

    num_boxes = 0
    for batch_idx in range(batch_size):
        num_boxes = num_boxes + len(preds[batch_idx])

    bounding_boxes = np.empty(shape=(num_boxes, 4), dtype=np.float)
    scores = np.empty(shape=(num_boxes), dtype=np.float)
    classes = np.empty(shape=(num_boxes), dtype=np.uint8)

    idx = 0
    for batch_idx in range(batch_size):
        i, j = batch_idx//side, batch_idx%side

        for pred in preds[batch_idx]:
            bounding_boxes[idx, 0] = (pred['y1'] + i)*tile_height
            bounding_boxes[idx, 1] = (pred['x1'] + j)*tile_width
            bounding_boxes[idx, 2] = (pred['y2'] + i)*tile_height
            bounding_boxes[idx, 3] = (pred['x2'] + j)*tile_width
            scores[idx] = pred['score']
            classes[idx] = pred['class']
            idx = idx+1

    return bounding_boxes, scores, classes



"""

These tools are used for the day-to-day management of geographic and tabular data.

copy_to_data_store copies data to your ArcGIS Data Store and creates a layer in your web GIS.
"""
import json as _json
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from ._util import _id_generator, _feature_input, _set_context, _create_output_service

_log = _logging.getLogger(__name__)

_use_async = True

def describe_data(input_layer,
                  extent_output=False,
                  sample_size=None,
                  output_name=None,
                  gis=None):
    """
    The Describe Dataset task provides an overview of your big data. The tool outputs a JSON
    detailing the input layer's geometry and time settings, schema, and summary statistics for each
    field. Optionally, the tool can output two types of descriptive feature layers: a feature layer
    representing a sample of your input features, and a single polygon feature layer that
    represents the extent of the input features. You can choose to output one or both.

        Usage Notes:

        Only available at ArcGIS Enterprise 10.7 and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer. The point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    extent_output     Optional Boolean. The task will output a single rectangle
                      feature representing the extent of the input_layer if this value
                      is set to true. The default is False.
    ----------------  ---------------------------------------------------------------
    sample_size       Optional integer. The task will output a feature layer
                      representing a sample of features from the input_layer. Specify
                      the number of sample features to return. If the input value is
                      0 or empty then no sample layer will be created. The output
                      will have the same schema, geometry, and time type as the input
                      layer. The default is None.
    ----------------  ---------------------------------------------------------------
    output_name       optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ================  ===============================================================

    :returns: FeatureLayer

    """
    kwargs = locals()
    tool_name = "DescribeDataset"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Describe_Dataset_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Merge Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "extent_output" : (bool, "extentOutput"),
        "sample_size" : (int, "sampleSize"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }

    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise
    return



def dissolve_boundaries(input_layer,
                        dissolve_fields=None,
                        summary_fields=None,
                        multipart=False,
                        output_name=None,
                        gis=None):
    """

    The Dissolve Boundaries task finds polygons that intersect or have the same field values and merges them together to form a single polygon.

    Examples:

        A city council wants to control liquor sales by refusing new licenses to stores within 1,000 feet of schools, libraries, and parks. After creating a 1,000-foot buffer around the schools, libraries, and parks, the buffered layers can be joined together and the boundaries can be dissolved to create a single layer of restricted areas.

    Usage Notes:



    Only available at ArcGIS Enterprise 10.7 and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer. The point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    dissolve_fields   Optional string. A comma seperated list of strings for each
                      field that you want to dissolve on.
    ----------------  ---------------------------------------------------------------
    summary_fields    Optional list. Calculate one or more statistics for the
                      dissolved areas by using the summary_fields parameter. The
                      input is a list of key/value pairs in the following format:

                         [{"statisticType" : "<stat>", "onStatisticField" : "<field name>"}]

                      Allows statistics are:

                        + Any (string fields only)
                        + Count
                        + Sum
                        + Minimum
                        + Maximum
                        + Average
                        + Variance
                        + Standard Deviation

                      Example:

                       summary_fields = [{"statisticType" : "Sum", "onStatisticField" : "quadrat_area_km2"},
                                         {"statisticType" : "Mean", "onStatisticField" : "soil_depth_cm"},
                                         {"statisticType" : "Any", "onStatisticField" : "quadrat_desc"}]
    ----------------  ---------------------------------------------------------------
    multipart         Optional boolean. If True, the output service can contain
                      multipart features. If False (default):, the output service
                      will only contain single-part features, and individual features
                      will be created for each part.
    ----------------  ---------------------------------------------------------------
    output_name       optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ================  ===============================================================

    :returns: FeatureLayer
    """
    kwargs = locals()
    tool_name = "DissolveBoundaries"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Dissolve_Bounds_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Merge Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "multipart": (bool, "multipart"),
        "summary_fields": (list, "summaryFields"),
        "dissolve_fields" : (list, "dissolveFields"),
        "output_name": (str, "OutputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }

    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise
    return

def merge_layers(input_layer, merge_layer, merge_attributes=None, output_name=None, gis=None):
    """
    The Merge Layers task combines two feature layers to create a single output layer. The tool
    requires that both layers have the same geometry type (tabular, point, line, or polygon). If
    time is enabled on one layer, the other must also be time enabled and have the same time type
    (instant or interval). The result will always contain all fields from the input layer. All
    fields from the merge layer will be included by default, or you can specify custom merge rules
    to define the resulting schema. For example:

    - I have three layers for England, Wales, and Scotland and I want a single layer of Great
      Britain. I can use Merge Layers to combine the areas and maintain all fields from each area.
    - I have two layers containing parcel information for contiguous townships. I want to join them
      together into a single layer, keeping only the fields that have the same name and type in the
      two layers.

    Only available at **ArcGIS Enterprise 10.7** and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       Required FeatureLayer. The point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    merge_layer       Required FeatureLayer. The point, line, or polygon features to
                      merge with the input_layer. The merge_layer must contain the
                      same geometry type (tabular, point, line, or polygon) and the
                      same time type (none, instant, or interval) as the input_layer.
                      All fields in the merge_layer will be included in the result
                      layer by default or you can define merge_attributes to
                      customize the resulting schema.
    ----------------  ---------------------------------------------------------------
    merge_attributes  Optional list. Defines how the fields in mergeLayer will be
                      modified. By default, all fields from both inputs will be
                      included in the output layer.

                      If a field exists in one layer but not the other, the output
                      layer will still contain the field. The output field will
                      contain null values for the input features that did not have the
                      field. For example, if the input_layer contains a field named
                      TYPE but the merge_layer does not contain TYPE, the output will
                      contain TYPE, but its values will be null for all the features
                      copied from the merge_layer.

                      You can control how fields in the merge_layer are written to the
                      output layer using the following merge types that operate on a
                      specified merge_layer field:

                      + Remove - The field in the merge_layer will be removed from the output layer.
                      + Rename - The field in the merge_layer will be renamed in the output layer. You cannot rename a field in the merge_layer to a field in the inputLayer. If you want to make field names equivalent, use Match.
                      + Match - A field in the merge_layer is made equivalent to a field in the input_layer specified by mergeValue. For example, the input_layer has a field named CODE and the merge_layer has a field named STATUS. You can match STATUS to CODE, and the output will contain the CODE field with values of the STATUS field used for features copied from the merge_layer. Type casting is supported (for example, double to integer, integer to string) except for string to numeric.
                      REST web example:

                      Syntax: This example matches Average_Sales to Mean_Sales,
                              removesBonus, and renamesField4 to Errors.

                      ```

                        [{
                            "mergeLayerField": "Mean_Sales",
                            "mergeType": "Match",
                            "mergeValue": "Average_Sales"
                        },
                        {
                            "mergeLayerField": "Bonus",
                            "mergeType": "Remove",
                        },
                        {
                            "mergeLayerField": "Field4",
                            "mergeType": "Rename",
                            "mergeValue": "Errors"
                        }]

                      ```

    ----------------  ---------------------------------------------------------------
    output_name       Optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               Optional GIS. The GIS object where the analysis will take place.
    ================  ===============================================================

    :returns: FeatureLayer
    """
    kwargs = locals()
    tool_name = "MergeLayers"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():

        if value is not None:
            params[key] = value
        elif key == 'merge_attributes' and value is None:
            params[key] = []
    if output_name is None:
        output_service_name = 'Merge_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Merge Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "merge_layer": (_FeatureSet, "mergeLayer"),
        "merge_attributes" : (list, "mergingAttributes"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return


def clip_layer(input_layer, clip_layer, output_name=None, gis=None):
    """
    Clip_layer features from one layer to the extent of a boundary layer. Use this tool to cut out a piece
    of one feature class using one or more of the features in another feature class as a cookie
    cutter. This is particularly useful for creating a new feature layers - also referred to as study
    area or area of interest (AOI)- that contains a geographic subset of the features in another,
    larger feature class.

    Only available at **ArcGIS Enterprise 10.7** and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer. The point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    clip_layer        required FeatureLayer. The features that will be clipping the input_layer features.
    ----------------  ---------------------------------------------------------------
    output_name       optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ================  ===============================================================

    :returns: FeatureLayer

    """
    kwargs = locals()
    tool_name = "ClipLayer"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value
    if output_name is None:
        output_service_name = 'Clip_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Overlay Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "clip_layer": (_FeatureSet, "clipLayer"),
        "outputType" : (str, 'outputType'),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return


def overlay_data(input_layer, overlay_layer, overlay_type="intersect", output_name=None, gis=None):
    """
    Only available at ArcGIS Enterprise 10.6.1 and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer. The point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    overlay_layer     required FeatureLayer. The features that will be overlaid with the input_layer features.
    ----------------  ---------------------------------------------------------------
    overlay_type      optional string. The type of overlay to be performed.
                      Values: intersect, erase

                      + intersect - Computes a geometric intersection of the input layers. Features or portions of features that overlap in both the inputLayer and overlayLayer layers will be written to the output layer. This is the default.
                      + erase - Only those features or portions of features in the overlay_layer that are not within the features in the input_layer layer are written to the output.
    ----------------  ---------------------------------------------------------------
    output_name       optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ================  ===============================================================

    :returns: FeatureLayer
    """
    kwargs = locals()
    tool_name = "OverlayLayers"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
        "outputType" : "Input",
        'tolerance' : 0,
        'snapToInput' : 'false'
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Overlay_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Overlay Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "overlay_layer": (_FeatureSet, "overlayLayer"),
        "outputType" : (str, 'outputType'),
        "overlay_type" : (str, "overlayType"),
        "output_name": (str, "OutputName"),
        "context": (str, "context"),
        'tolerance' : (int, 'tolerance'),
        "output": (_FeatureSet, "output"),
        'snapToInput' : (str, 'snapToInput')
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return


def append_data(input_layer, append_layer, field_mapping=None, gis=None):
    """
    Only available at ArcGIS Enterprise 10.6.1 and later.

    The Append Data task appends tabular, point, line, or polygon data to an existing layer.
    The input layer must be a hosted feature layer. The tool will add the appended data as
    rows to the input layer. No new output layer is created.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer , The table, point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    append_layer      required FeatureLayer. The table, point, line, or polygon features
                      to be appended to the input_layer. To append geometry, the
                      append_layer must have the same geometry type as the
                      input_layer. If the geometry types are not the same, the
                      append_layer geometry will be removed and all other matching
                      fields will be appended. The geometry of the input_layer will
                      always be maintained.
    ----------------  ---------------------------------------------------------------
    field_mapping     Defines how the fields in append_layer are appended to the
                      input_layer.

                      The following are set by default:

                        - All append_layer fields that match input_layer schema will be appended.
                        - Fields that exist in the input_layer and not in the append_layer will be appended with null values.
                        - Fields that exist in the append_layer and not in the input_layer will not be appended.

                      Optionally choose how input_layer fields will be appended from the following:

                      - AppendField - Matches the input_layer field with an append_layer field of a different name. Field types must match.
                      - Expression - Calculates values for the resulting field. Values are calculated using Arcade expressions. To assign null values, use 'null'.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS, the GIS on which this tool runs. If not
                      specified, the active GIS is used.
    ================  ===============================================================

    :returns: boolean

    """
    kwargs = locals()
    tool_name = "AppendData"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "append_layer": (_FeatureSet, "appendLayer"),
        "field_mapping" : (str, "fieldMapping"),
        "context": (str, "context")
    }
    return_values = [
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return True
    except:
        raise

    return False


def calculate_fields(input_layer,
                     field_name,
                     data_type,
                     expression,
                     track_aware=False,
                     track_fields=None,
                     time_boundary_split=None,
                     time_split_unit=None,
                     time_reference=None,
                     output_name=None,
                     gis=None
                     ):
    """
    The Calculate Field task works with a layer to create and populate a
    new field. The output is a new feature layer, that is the same as the
    input features, with the additional field added.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  required service , The table, point, line or polygon features
                                 containing potential incidents.
    --------------------------   ---------------------------------------------------------------
    field_name                   required string, A string representing the name of the new
                                 field. If the name already exists in the dataset, then a
                                 numeric value will be appended to the field name.
    --------------------------   ---------------------------------------------------------------
    data_type                    required string, the type for the new field.
                                 Values: Date, Double, Integer, String
    --------------------------   ---------------------------------------------------------------
    expression                   required string, An Arcade expression used to calculate the new
                                 field values. You can use any of the Date, Logical,
                                 Mathematical or Text function available with Arcade.
    --------------------------   ---------------------------------------------------------------
    track_aware                  optional boolean, Boolean value denoting if the expression is
                                 track aware.
                                 Default: False
    --------------------------   ---------------------------------------------------------------
    track_fields                 optional string, The fields used to identify distinct tracks.
                                 There can be multiple track_fields. track_fields are only
                                 required when track_aware is true.
    --------------------------   ---------------------------------------------------------------
    time_boundary_split          Optional Int.  A time boundary to detect and incident.
    --------------------------   ---------------------------------------------------------------
    time_split_unit              Optional String.  The unit to detect an incident is `time_boundary_split` is used.
                                 Allowed values: Years, Months, Weeks, Days, Hours, Minutes, Seconds, Milliseconds
    --------------------------   ---------------------------------------------------------------
    time_reference               Optional Datetime. The starting date/time where analysis will
                                 begin from.
    --------------------------   ---------------------------------------------------------------
    output_name                  optional string, The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    gis                          optional GIS, the GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    ==========================  ================================================================

    :returns:
       Feature Layer
    """
    kwargs = locals()
    tool_name = "CalculateField"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Calculate_Fields_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Calculate Fields')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "field_name" : (str, "fieldName"),
        "data_type" : (str, "dataType"),
        "expression" : (str, "expression"),
        "track_aware" : (bool, "trackAware"),
        "track_fields" : (str, "trackFields"),
        "time_boundary_split" : (int, "timeBoundarySplit"),
        "time_split_unit" : (str, "timeBoundarySplitUnit"),
        "time_reference" : (datetime.datetime, "timeBoundaryReference"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return

def copy_to_data_store(
    input_layer,
    output_name = None,
    gis = None):
    """

    Copies an input feature layer or table to an ArcGIS Data Store and creates a layer in your web GIS.

    For example

    * Copy a collection of .csv files in a big data file share to the spatiotemporal data store for visualization.

    * Copy the features in the current map extent that are stored in the spatiotemporal data store to the relational data store.

    This tool will take an input layer and copy it to a data store. Data will be copied to the ArcGIS Data Store and will be stored in your relational or spatiotemporal data store.

    For example, you could copy features that are stored in a big data file share to a relational data store and specify that only features within the current map extent will be copied. This would create a hosted feature service with only those features that were within the specified map extent.

   Parameters:

   input_layer: Input Layer (feature layer). Required parameter.

   output_name: Output Layer Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Layer as a feature layer collection item


    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Data Store Copy_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Copy To Data Store')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Layer"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Layer", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, "CopyToDataStore", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

copy_to_data_store.__annotations__ = {
    'output_name': str}








"""
Provides class, methods and functions to manage resources for a given GIS.
"""
import os
import json
import tempfile


class PortalResourceManager(object):
    """
    Helper class to manage a GIS' resources

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    gis               required GIS, connection to ArcGIS Online or ArcGIS Enterprise
    ================  ===============================================================


    """
    _workdir = None

    def __init__(self, gis):
        """Creates helper object to manage custom roles in the GIS"""
        self._gis = gis
        self._portal = gis._portal
        self._is_portal = self._gis.properties.isPortal
        self._workdir = tempfile.gettempdir()


    def add(self, key=None, path=None, text=None, **kwargs):
        """
        The add resource operation allows the administrator to add a file
        resource, for example, the organization's logo or custom banner.
        The resource can be used by any member of the organization. File
        resources use storage space from your quota and are scanned for
        viruses.


        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        key               optional string, look up key for file
        ----------------  ---------------------------------------------------------------
        path              optional string, file path to the local resource to upload
        ----------------  ---------------------------------------------------------------
        text              optional string, text value to add to the site's resources
        ----------------  ---------------------------------------------------------------
        access            optional string, sets the access level for the resource. The
                          default is 'public'.
                          Values: public, org, orgprivate
        ================  ===============================================================

        :returns:
           boolean
        """
        access = kwargs.pop("access", None)
        files = None
        if key is None and path:
            key = os.path.basename(path)
        elif key is None and path is None:
            raise ValueError("key must be populated is path is null")
        url = "portals/self/addresource"
        postdata = {
            "f" : "json",
            "key" : key,
        }
        if path:
            files = {
                'file' : path
            }
        if text:
            if isinstance(text, dict):
                postdata['text'] = json.dumps(text)
            elif isinstance(text, str):
                from arcgis._impl.common._utils import _to_utf8
                postdata['text'] = _to_utf8(text)
        else:
            if self._portal.is_arcgisonline == False:
                postdata['text'] = ""
        if self._is_portal == False:
            url = "portals/%s/addResource" % self._gis.properties.id
            if text is None:
                postdata['text'] = ""
            if access:
                postdata['access'] = access
            else:
                postdata['access'] = 'public'

        resp = self._portal.con.post(url,
                                     postdata, files=files)
        if 'success' in resp:
            return resp['success']
        return resp

    def delete(self, key):
        """
        The Remove Resource operation allows the administrator to remove
        a file resource.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        key               optional string, look up key for file to delete
        ================  ===============================================================

        :returns:
           boolean
        """
        postdata = {
            "f" : "json",
            "key" : key,
            }
        resp = self._portal.con.post('portals/self/removeresource',
                                     postdata)
        if 'success' in resp:
            return resp['success']
        return resp

    #----------------------------------------------------------------------
    def list(self, start=1, num=100):
        """
        returns a list of resources uploaded to portal.  The items can be
        images, files and other content used to stylize and modify a
        portal's appearance.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        start             optional int, start location of the search. The default is
                          a value of 1
        ----------------  ---------------------------------------------------------------
        num               optional int, the number of search results to return at one
                          time. The value ranges between 1-100 (max).
                          Default: 100
        ================  ===============================================================

        :returns:
           boolean
        """
        postdata = {
            "f" : "json",
            'start' : start,
            'num' : num
        }
        resp = self._portal.con.post('portals/self/resources',
                                     postdata)
        if 'resources' in resp:
            return resp['resources']
        return resp

    def get(self, resource_name, download_path=None):
        """
        Download or get a portal resource item

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        resource_name     optional string, key/name of data
        ----------------  ---------------------------------------------------------------
        download_path     optional string, save folder location
        ================  ===============================================================

        :return:
           path to data or raw data if not file.
        """
        data_path = 'portals/self/resources/' + resource_name
        if not download_path:
            download_path = self._workdir

        download_path = self._portal.con.get(path=data_path,
                                             file_name=resource_name,
                                             out_folder=download_path,
                                             try_json=False,
                                             force_bytes=False)
        return download_path



"""
contains operations to read/write geodataframe object
to and from disk.
"""
from __future__ import print_function
from __future__ import division
from .fileops import from_featureclass, to_featureclass
from .sqliteio import to_sqlite
from .serviceops import from_layer



"""
These functions help you answer one of the most common questions posed in spatial analysis: "What is near what?"

connect_origins_to_destinations measures the travel time or distance between pairs of points.
create_buffers create areas of equal distance from features.
create_drive_time_areas finds areas around locations that can be reached within a time period.
find_nearest identifies those places that are the closest to known locations.
plan_routes determines the best way to route a fleet of vehicles to visit many stops.
"""
import arcgis as _arcgis
from arcgis._impl.common._utils import _date_handler

def connect_origins_to_destinations(
        origins_layer,
        destinations_layer,
        measurement_type="DrivingTime",
        origins_layer_route_id_field=None,
        destinations_layer_route_id_field=None,
        time_of_day=None,
        time_zone_for_time_of_day="GeoLocal",
        output_name=None,
        context=None,
        gis=None):
    """
    Calculates routes between pairs of points.

    Parameters
    ----------
    origins_layer : Required layer (see Feature Input in documentation)
        The routes start from points in the origins layer.
    destinations_layer : Required layer (see Feature Input in documentation)
        The routes end at points in the destinations layer.
    measurement_type : Required string
        The routes can be determined by measuring travel distance or travel time along street network using different
        travel modes or by measuring straight line distance.
    origins_layer_route_id_field : Optional string
        The field in the origins layer containing the IDs that are used to match an origin with a destination.
    destinations_layer_route_id_field : Optional string
        The field in the destinations layer containing the IDs that are used to match an origin with a destination.
    time_of_day : Optional datetime.datetime
        When measurementType is DrivingTime, this value specifies the time of day to be used for driving time
        calculations based on traffic. WalkingTime and TruckingTime measurementType do not support calculations
        based on traffic.
    time_zone_for_time_of_day : Optional string
        Determines if the value specified for timeOfDay is specified in UTC or in a time zone that is local to the
        location of the origins.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    dict with the following keys:
       "routes_layer" : layer (FeatureCollection)
       "unassigned_origins_layer" : layer (FeatureCollection)
       "unassigned_destinations_layer" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.connect_origins_to_destinations(
        origins_layer,
        destinations_layer,
        measurement_type,
        origins_layer_route_id_field,
        destinations_layer_route_id_field,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context)


def create_buffers(
        input_layer,
        distances=[],
        field=None,
        units="Meters",
        dissolve_type="None",
        ring_type="Disks",
        side_type="Full",
        end_type="Round",
        output_name=None,
        context=None,
        gis=None):
    """
    Creates buffer polygon(s) around input features.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The input to be buffered.
    distances : Optional list of floats
        The distance(s) that will be buffered.
    field : Optional string
        Buffers will be created using field values.
    units : Optional string
        The linear unit to be used with the distance value(s).
    dissolve_type : Optional string
        Specifies the dissolve to be performed to remove buffer overlap.
    ring_type : Optional string
        The ring type.
    side_type : Optional string
        The side(s) of the input that will be buffered.
    end_type : Optional string
        The shape of the buffer at the end of buffered line features.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    buffer_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_buffers(
        input_layer,
        distances,
        field,
        units,
        dissolve_type,
        ring_type,
        side_type,
        end_type,
        output_name,
        context)


def create_drive_time_areas(
        input_layer,
        break_values=[5, 10, 15],
        break_units="Minutes",
        travel_mode="Driving",
        overlap_policy="Overlap",
        time_of_day=None,
        time_zone_for_time_of_day="GeoLocal",
        output_name=None,
        context=None,
        gis=None):
    """


    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)

    break_values : Optional list of floats

    break_units : Optional string

    travel_mode : Optional string

    overlap_policy : Optional string

    time_of_day : Optional datetime.datetime

    time_zone_for_time_of_day : Optional string

    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    drive_time_areas_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_drive_time_areas(
        input_layer,
        break_values,
        break_units,
        travel_mode,
        overlap_policy,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context)


def find_nearest(
        analysis_layer,
        near_layer,
        measurement_type="StraightLine",
        max_count=100,
        search_cutoff=2147483647,
        search_cutoff_units=None,
        time_of_day=None,
        time_zone_for_time_of_day="GeoLocal",
        output_name=None,
        context=None,
        gis=None):
    """
    Measures the straight-line distance, driving distance, or driving time from features in the analysis layer to
    features in the near layer, and copies the nearest features in the near layer to a new layer. Returns a layer
    containing the nearest features and a line layer that links the start locations to their nearest locations.

    Parameters
    ----------
    analysis_layer : Required layer (see Feature Input in documentation)
        For each feature in this layer, the task finds the nearest features from the nearLayer.
    near_layer : Required layer (see Feature Input in documentation)
        The features from which the nearest locations are found.
    measurement_type : Required string
        The nearest locations can be determined by measuring straight-line distance, driving distance, or driving time
    max_count : Optional int
        The maximum number of near locations to find for each feature in analysisLayer.
    search_cutoff : Optional float
        Limits the search range to this value
    search_cutoff_units : Optional string
        The units for the value specified as searchCutoff
    time_of_day : Optional datetime.datetime
        When measurementType is DrivingTime, this value specifies the time of day to be used for driving time
        calculations based on traffic.
    time_zone_for_time_of_day : Optional string

    output_name : Optional string
        Additional properties such as output feature service name
    context : Optional string
        Additional settings such as processing extent and output spatial reference
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    dict with the following keys:
       "nearest_layer" : layer (FeatureCollection)
       "connecting_lines_layer" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_nearest(
        analysis_layer,
        near_layer,
        measurement_type,
        max_count,
        search_cutoff,
        search_cutoff_units,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context)


def plan_routes(
        stops_layer,
        route_count,
        max_stops_per_route,
        route_start_time,
        start_layer,
        start_layer_route_id_field=None,
        return_to_start=True,
        end_layer=None,
        end_layer_route_id_field=None,
        travel_mode="Driving",
        stop_service_time=0,
        max_route_time=525600,
        include_route_layers=False,
        output_name=None,
        context=None,
        gis=None):
    """
    You provide a set of stops and the number of vehicles available to visit the stops, and Plan Routes determines how
    to efficiently assign the stops to the vehicles and route the vehicles to the stops.

    Use this tool to plan work for a mobile team of inspectors, appraisers, in-home support service providers, and
    others; deliver or pick up items from remote locations; or offer transportation services to people.

    Parameters
    ----------

    stops_layer : Required layer (see Feature Input in documentation)

    route_count : Required int

    max_stops_per_route : Required int

    route_start_time : Required datetime.datetime

    start_layer : Required layer (see Feature Input in documentation)

    start_layer_route_id_field : Optional string

    return_to_start : Optional bool

    end_layer : Optional layer (see Feature Input in documentation)

    end_layer_route_id_field : Optional string

    travel_mode : Optional string

    stop_service_time : Optional float

    max_route_time : Optional float

    include_route_layers : Optional bool

    output_name : Optional string

    context : Optional string

    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


    Returns
    -------
    dict with the following keys:
       "routes_layer" : layer (FeatureCollection)
       "assigned_stops_layer" : layer (FeatureCollection)
       "unassigned_stops_layer" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.plan_routes(
        stops_layer,
        route_count,
        max_stops_per_route,
        _date_handler(route_start_time),
        start_layer,
        start_layer_route_id_field,
        return_to_start,
        end_layer,
        end_layer_route_id_field,
        travel_mode,
        stop_service_time,
        max_route_time,
        include_route_layers,
        output_name,
        context)



try:
    import torch
    from fastai.vision.learner import create_cnn
    from fastai.callbacks.hooks import model_sizes
    from fastai.vision.learner import create_body
    from torchvision.models import resnet34
    import numpy as np
    from ._ssd_utils import SSDHead, BCE_Loss, FocalLoss, one_hot_embedding, nms
    from .._data import prepare_data
    import json
    import os
    from pathlib import Path
    from ._codetemplate import code
    HAS_FASTAI = True
except Exception as e:
    HAS_FASTAI = False

def _raise_fastai_import_error():
    raise Exception('This module requires fastai, PyTorch and torchvision as its dependencies. Install it using "conda install -c pytorch -c fastai fastai pytorch torchvision"')

_EMD_TEMPLATE = {
    "Framework": "arcgis.learn.models._inferencing",
    "InferenceFunction":"ArcGISObjectDetector.py",
    "ModelConfiguration": "_DynamicSSD",
    "ModelFile":"",
    "ModelType":"ObjectDetection",
    "ImageHeight":None,
    "ImageWidth":None,
    "ExtractBands":[0,1,2],
    "Grids":None,
    "Zooms":None,
    "Ratios":None,
    "Classes" : []
}

_CLASS_TEMPLATE = {
        "Value": 0,
        "Name": "Pool",
        "Color": [0, 255, 0]
      }

class _EmptyData():
    def __init__(self, path, c, loss_func, chip_size):
        self.path = path
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        self.c = c
        self.loss_func = loss_func
        self.chip_size = chip_size

class SingleShotDetector(object):

    """
    Creates a Single Shot Detector with the specified grid sizes, zoom scales
    and aspect  ratios. Based on Fast.ai MOOC Version2 Lesson 9.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    grids                   Required list. Grid sizes used for creating anchor
                            boxes.
    ---------------------   -------------------------------------------
    zooms                   Optional list. Zooms of anchor boxes.
    ---------------------   -------------------------------------------
    ratios                  Optional list of tuples. Aspect ratios of anchor
                            boxes.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `SingleShotDetector`, which
                            is `resnet34` by default.
    ---------------------   -------------------------------------------
    dropout                 Optional float. Dropout propbability. Increase it to
                            reduce overfitting.
    ---------------------   -------------------------------------------
    bias                    Optional float. Bias for SSD head.
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. Uses Focal Loss if True.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :returns: `SingleShotDetector` Object
    """

    def __init__(self, data, grids=[4, 2, 1], zooms=[0.7, 1., 1.3], ratios=[[1., 1.], [1., 0.5], [0.5, 1.]],
                 backbone=None, drop=0.3, bias=-4., focal_loss=False, pretrained_path=None):

        super().__init__()

        self._device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

        if not HAS_FASTAI:
            _raise_fastai_import_error()

        if backbone is None:
            backbone = resnet34

        self._create_anchors(grids, zooms, ratios)

        num_features = model_sizes(create_body(backbone), size=(data.chip_size, data.chip_size))[-1][-1]

        ssd_head = SSDHead(grids, self._anchors_per_cell, data.c, num_features=num_features, drop=drop, bias=bias)

        self._data = data
        self.learn = create_cnn(data=data, arch=backbone, custom_head=ssd_head)
        self.learn.model = self.learn.model.to(self._device)

        if pretrained_path is not None:
            self.load(pretrained_path)

        if focal_loss:
            self._loss_f = FocalLoss(data.c)
        else:
            self._loss_f = BCE_Loss(data.c)

        self.learn.loss_func = self._ssd_loss

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a Single Shot Detector from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================

        :returns: `SingleShotDetector` Object
        """
        emd_path = Path(emd_path)
        emd = json.load(open(emd_path))
        model_file = Path(emd['ModelFile'])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
        if data is None:
            empty_data = _EmptyData(path='str', loss_func=None, c=len(class_mapping) + 1, chip_size=emd['ImageHeight'])
            return cls(empty_data, emd['Grids'], emd['Zooms'], emd['Ratios'], pretrained_path=str(model_file))
        else:
            return cls(data, emd['Grids'], emd['Zooms'], emd['Ratios'], pretrained_path=str(model_file))


    def lr_find(self):
        """
        Runs the Learning Rate Finder, and displays the graph of it's output.
        Helps in choosing the optimum learning rate for training the model.
        """
        from IPython.display import clear_output
        self.learn.lr_find()
        clear_output()
        self.learn.recorder.plot()

    def fit(self, epochs=10, lr=slice(1e-4,3e-3)):
        """
        Train the model for the specified number of epocs and using the
        specified learning rates

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Required float or slice of floats. Learning rate
                                to be used for training the model. Select from
                                the `lr_find` plot.
        =====================   ===========================================
        """
        self.learn.fit(epochs, lr)

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the detector for fine-tuning.
        """
        self.learn.unfreeze()

    def _create_anchors(self, anc_grids, anc_zooms, anc_ratios):

        self.grids = anc_grids
        self.zooms = anc_zooms
        self.ratios =  anc_ratios

        anchor_scales = [(anz*i, anz*j) for anz in anc_zooms for (i,j) in anc_ratios]

        self._anchors_per_cell = len(anchor_scales)

        anc_offsets = [1/(o*2) for o in anc_grids]

        anc_x = np.concatenate([np.repeat(np.linspace(ao, 1-ao, ag), ag)
                                for ao,ag in zip(anc_offsets,anc_grids)])
        anc_y = np.concatenate([np.tile(np.linspace(ao, 1-ao, ag), ag)
                                for ao,ag in zip(anc_offsets,anc_grids)])
        anc_ctrs = np.repeat(np.stack([anc_x,anc_y], axis=1), self._anchors_per_cell, axis=0)

        anc_sizes  =   np.concatenate([np.array([[o/ag,p/ag] for i in range(ag*ag) for o,p in anchor_scales])
                       for ag in anc_grids])

        self._grid_sizes = torch.Tensor(np.concatenate([np.array([ 1/ag  for i in range(ag*ag) for o,p in anchor_scales])
                       for ag in anc_grids])).unsqueeze(1).to(self._device)

        self._anchors = torch.Tensor(np.concatenate([anc_ctrs, anc_sizes], axis=1)).float().to(self._device)

        self._anchor_cnr = self._hw2corners(self._anchors[:,:2], self._anchors[:,2:])

    def _hw2corners(self, ctr, hw):
        return torch.cat([ctr-hw/2, ctr+hw/2], dim=1)

    def _get_y(self, bbox, clas):
        bbox = bbox.view(-1,4) #/sz
        bb_keep = ((bbox[:,2]-bbox[:,0])>0).nonzero()[:,0]
        return bbox[bb_keep],clas[bb_keep]

    def _actn_to_bb(self, actn, anchors, grid_sizes):
        actn_bbs = torch.tanh(actn)
        actn_centers = (actn_bbs[:,:2]/2 * grid_sizes) + anchors[:,:2]
        actn_hw = (actn_bbs[:,2:]/2+1) * anchors[:,2:]
        return self._hw2corners(actn_centers, actn_hw)

    def _map_to_ground_truth(self, overlaps, print_it=False):
        prior_overlap, prior_idx = overlaps.max(1)
        if print_it: print(prior_overlap)
        gt_overlap, gt_idx = overlaps.max(0)
        gt_overlap[prior_idx] = 1.99
        for i,o in enumerate(prior_idx): gt_idx[o] = i
        return gt_overlap, gt_idx


    def _ssd_1_loss(self, b_c, b_bb, bbox, clas, print_it=False):
        bbox,clas = self._get_y(bbox,clas)
        bbox = self._normalize_bbox(bbox)

        a_ic = self._actn_to_bb(b_bb, self._anchors, self._grid_sizes)
        overlaps = self._jaccard(bbox.data, self._anchor_cnr.data)
        try:
            gt_overlap,gt_idx = self._map_to_ground_truth(overlaps,print_it)
        except Exception as e:
            return 0.,0.
        gt_clas = clas[gt_idx]
        pos = gt_overlap > 0.4
        pos_idx = torch.nonzero(pos)[:,0]
        gt_clas[1-pos] = 0
        gt_bbox = bbox[gt_idx]
        loc_loss = ((a_ic[pos_idx] - gt_bbox[pos_idx]).abs()).mean()
        clas_loss  = self._loss_f(b_c, gt_clas)
        return loc_loss, clas_loss

    def _ssd_loss(self, pred, targ1, targ2, print_it=False):
        lcs,lls = 0.,0.
        for b_c,b_bb,bbox,clas in zip(*pred, targ1, targ2):
            loc_loss,clas_loss = self._ssd_1_loss(b_c,b_bb,bbox.to(self._device),clas.to(self._device),print_it)
            lls += loc_loss
            lcs += clas_loss
        if print_it: print('loc: {lls}, clas: {lcs}'.format(lls=lls, lcs=lcs))
        return lls+lcs

    def _intersect(self,box_a, box_b):
        max_xy = torch.min(box_a[:, None, 2:], box_b[None, :, 2:])
        min_xy = torch.max(box_a[:, None, :2], box_b[None, :, :2])
        inter = torch.clamp((max_xy - min_xy), min=0)
        return inter[:, :, 0] * inter[:, :, 1]

    def _box_sz(self, b):
        return ((b[:, 2]-b[:, 0]) * (b[:, 3]-b[:, 1]))

    def _jaccard(self, box_a, box_b):
        inter = self._intersect(box_a, box_b)
        union = self._box_sz(box_a).unsqueeze(1) + self._box_sz(box_b).unsqueeze(0) - inter
        return inter / union

    def _normalize_bbox(self, bbox):
        return (bbox+1.)/2.

    def _create_zip(self, zipname, path):
        import shutil
        zip_file = shutil.make_archive(zipname, 'zip', path)
        shutil.move(zip_file, path)

    def _create_emd(self, path):
        import random
        _EMD_TEMPLATE['ModelFile'] = path.name
        _EMD_TEMPLATE['ImageHeight'] = self._data.chip_size
        _EMD_TEMPLATE['ImageWidth'] = self._data.chip_size
        _EMD_TEMPLATE['Grids'] = self.grids
        _EMD_TEMPLATE['Zooms'] = self.zooms
        _EMD_TEMPLATE['Ratios'] = self.ratios
        for i, class_name in enumerate(self._data.classes[1:]): # 0th index is background
            _CLASS_TEMPLATE["Value"] = i
            _CLASS_TEMPLATE["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            _CLASS_TEMPLATE["Color"] = color
            _EMD_TEMPLATE['Classes'].append(_CLASS_TEMPLATE.copy())

        json.dump(_EMD_TEMPLATE, open(path.with_suffix('.emd'), 'w'), indent=4)
        return path.stem

    def save(self, name_or_path):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro

        Train the model for the specified number of epocs and using the
        specified learning rates

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name. and creates
                                all the intermediate directories.
        =====================   ===========================================
        """
        if '\\' in name_or_path or '/' in name_or_path:
            path = Path(name_or_path)
            name = path.parts[-1]
            # to make fastai save to both path and with name
            temp = self.learn.path
            self.learn.path = path
            self.learn.model_dir = ''
            if not os.path.exists(self.learn.path):
                os.makedirs(self.learn.path)
            saved_path = self.learn.save(name, return_path=True)
            # undoing changes to self.learn.path and self.learn.model
            self.learn.path = temp
            self.learn.model_dir = 'models'
        else:
            temp = self.learn.path
            # fixing fastai bug
            self.learn.path = self.learn.path.parent
            self.learn.model_dir =  Path(self.learn.model_dir) /  name_or_path
            if not os.path.exists(self.learn.path / self.learn.model_dir):
                os.makedirs(self.learn.path / self.learn.model_dir)
            saved_path = self.learn.save(name_or_path,  return_path=True)
            # undoing changes to self.learn.path
            self.learn.path = temp
            self.learn.model_dir = 'models'

        zip_name = self._create_emd(saved_path)
        with open(saved_path.parent / _EMD_TEMPLATE['InferenceFunction'], 'w') as f:
            f.write(code)
        self._create_zip(zip_name, str(saved_path.parent))
        print('Created model files at {spp}'.format(spp=saved_path.parent))


    def load(self, name_or_path):
        """
        Loads a saved model for inferencing or fine tuning from the specified
        path or model name.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to load from
                                the pre-defined location. If path is passed then
                                it loads from the specified path with model name
                                as directory name. Path to ".pth" file can also
                                be passed
        =====================   ===========================================
        """
        if '\\' in name_or_path or '/' in name_or_path:
            path = Path(name_or_path)
            # to make fastai from both path and with name
            temp = self.learn.path
            if path.is_file():
                name = path.stem
                self.learn.path = path.parent
            else:
                name = path.parts[-1]
                self.learn.path = path
            self.learn.model_dir = ''
            self.learn.load(name)
            # undoing changes to self.learn.path and self.learn.model_dir
            self.learn.path = temp
            self.learn.model_dir = 'models'
        else:
            temp = self.learn.path
            # fixing fastai bug
            self.learn.path = self.learn.path.parent
            self.learn.model_dir =  Path(self.learn.model_dir) /  name_or_path
            self.learn.load(name_or_path)
            # undoing changes to self.learn.path
            self.learn.path = temp
            self.learn.model_dir = 'models'

    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self.learn.show_results(rows=rows, thresh=thresh, nms_overlap=nms_overlap, ssd=self)






"""
class to work with the living atlas
"""
from ..._impl.connection import _ArcGISConnection
from ..._impl.common._mixins import PropertyMap
from ...gis import GIS
from ...gis import Group, GroupManager
from ._base import BasePortalAdmin
########################################################################
class LivingAtlas(BasePortalAdmin):
    """
    Living Atlas of the World content is a collection of authoritative,
    ready-to-use, global geographic content available from ArcGIS Online.
    The content includes valuable maps, data layers, tools, services and
    apps for geographic analysis.
    When you make Living Atlas content available to your portal members,
    you're providing them with ready-made content that they can use
    alone or in combination with their own content to create maps,
    scenes, and apps and perform analysis in the portal map viewer or
    Insights for ArcGIS.

    :Note:
       Your portal must have access to the Internet to use Living Atlas
       content from ArcGIS Online

    Types of content available
    All the Living Atlas content you access from Portal for ArcGIS was
    created by Esri. If your portal can connect to the Internet, the
    following three levels of Living Atlas content are available to you from
    ArcGIS Online:
      1). Content that does not require you to sign in to an ArcGIS Online
          account
          - This content is available by default in Portal for ArcGIS.
      2). Subscriber content
          - Subscriber content is the collection of ready-to-use map layers,
            analytic tools, and services published by Esri that requires an
            ArcGIS Online organizational subscription account to access. This
            includes layers from Esri such as Landsat 8 imagery, NAIP imagery,
            landscape analysis layers, and historical maps. Subscriber content
            is provided as part of your organizational subscription and does
            not consume any credits. Layers included in the Living Atlas
            subscriber content are suitable for use with analysis tools.
      3). Premium content
         - Premium content is a type of subscriber content that requires an
         ArcGIS Online organizational subscription account to access and
         consumes credits. Access and credit information is listed in the
         description details for each item.
         Premium content provides portal members with access to ready-to-use
         content such as demographic and lifestyle maps as well as tools for
         geocoding, geoenrichment, network analysis, elevation analysis, and
         spatial analysis.


    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    url                 required string, the web address of the site to
                        manage licenses.
                        example:
                        https://<org url>/<wa>/portaladmin/system/content/livingatlas
    ---------------     ----------------------------------------------------
    gis                 required GIS, the gis connection object.
    ===============     ====================================================
    """
    _groupquery = None
    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None
    _living_atlas_group = None
    _living_atlas_content_group = None
    _groups = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""

        super(LivingAtlas, self).__init__(url=url, gis=gis)
        self._url = url.replace("http://", "https://")
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")

        self._init()
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        """initializer"""
        try:
            self._groupquery = self._gis.properties['livingAtlasGroupQuery']
        except:
            self._groupquery = 'title:"Living Atlas" AND owner:esri_livingatlas'
        groups = self._gis.groups
        self._groups = []
        for group in groups.search(query=self._groupquery):
            if group.title.lower() == "living atlas".lower():
                self._living_atlas_group =  group
            elif group.title.lower() == 'Living Atlas Analysis Layers'.lower():
                self._living_atlas_content_group = group
            self._groups.append(group)
            del group
        del groups
        self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def enable_public_access(self):
        """
        Enables the Public Living Atlas content.

        Living Atlas of the World content is a collection of authoritative,
        ready-to-use, global geographic content available from ArcGIS Online.
        The content includes valuable maps, data layers, tools, services and
        apps for geographic analysis.

        :returns:
           boolean. True means enabled, False means failure to enable.

        """
        url = self._url + "/share"
        results = []
        for g in self.groups:
            params = {
                "f" : "json",
                "groupId" : g.id,
                "type" : "Public"
            }
            res = self._con.post(path=url, postdata=params)
            results.append(res['status'] == 'success')
        return all(results)
    #----------------------------------------------------------------------
    def disable_public_access(self):
        """
        Disables the Public Living Atlas content.

        :returns:
           boolean. True means enabled, False means failure to enable.

        """
        url = self._url + "/unshare"
        results = []
        for g in self.groups:
            params = {
                "f" : "json",
                "groupId" : g.id,
                "type" : "Public"
            }
            res = self._con.post(path=url, postdata=params)
            results.append(res['status'] == 'success')
        return all(results)
    #----------------------------------------------------------------------
    def status(self, group):
        """
        returns the information about the sharing status of the Living
        Atlas

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        group               required string or Group object
        ===============     ====================================================
        """
        url = "%s/status" % self._url
        params = {"f" : "json"}
        if isinstance(group, str):
            params["groupId" ] = group
        elif isinstance(group, Group):
            params['groupId'] = group.id
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def upgrade(self):
        """
        Upgrades the Living Atlas Group to the latest version

        :return: Boolean
        """
        url = "%s/upgrade"
        params = {'f': 'json'}
        try:
            for g in self.groups:
                params['groupId'] = g.id
                self._con.post(url, params)
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def update_subscriber_account(self, username, password):
        """
        Updates the Username/Password for the Living Atlas Subscriber User.
        The account must be an ArcGIS Online account.

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            Required string. The user who will be used for
                            to access the subscriber Living Atlas content.
        ---------------     ----------------------------------------------------
        password            Required string. The credentials for the user above.
        ===============     ====================================================

        :return: boolean

        """
        url = "%s/update" % self._url
        r = []
        for g in self.groups:
            params = {
                "f" : "json",
                "groupId" : g.id,
                "type" : 'Premium',
                "username" : username,
                "password" : password
            }
            res = self._con.post(url, params)
            if 'success' in res:
                r.append(True)
            else:
                r.append(False)
        return all(r)
    #----------------------------------------------------------------------
    def update_premium_account(self, username, password):
        """
        Updates the Username/Password for the Living Atlas Premium User.
        The account must be an ArcGIS Online account.

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            Required string. The user who will be used for
                            to access the subscriber Living Atlas content.
        ---------------     ----------------------------------------------------
        password            Required string. The credentials for the user above.
        ===============     ====================================================

        :return: boolean

        """
        url = "%s/update" % self._url
        r = []
        for g in self.groups:
            params = {
                "f" : "json",
                "groupId" : g.id,
                "type" : 'Premium',
                "username" : username,
                "password" : password
            }
            res = self._con.post(url, params)
            if 'success' in res:
                r.append(True)
            else:
                r.append(False)
        return all(r)

    #----------------------------------------------------------------------
    @property
    def groups(self):
        """returns a list of all living atlas groups"""
        if self._groups is None:
            self._init()
        return self._groups
    #----------------------------------------------------------------------
    def validate_credentials(self, username, password, online_url=None):
        """
        returns the information about the sharing status of the Living
        Atlas

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            required string, username for AGOL
        ---------------     ----------------------------------------------------
        password            required string, login password for AGOL account
        ---------------     ----------------------------------------------------
        online_url          optional string, Url to ArcGIS Online site.
                            default is https://www.arcgis.com
        ===============     ====================================================

        :returns:
          boolean
        """
        if online_url is None:
            online_url = "https://www.arcgis.com"
        url = "%s/validate" % self._url
        params = {
            "username" : username,
            "password" : password,
            "onlineUrl" : online_url,
            "f" : "json",
        }
        res = self._con.post(path=url, postdata=params)
        return res['status'] == 'success'
    #----------------------------------------------------------------------
    def enable_premium_atlas(self, username, password):
        """
        Enables the Premium Livinng Atlas Content for a local portal.

        Premium content is a type of subscriber content that requires an
        ArcGIS Online organizational subscription account to access and
        consumes credits. Access and credit information is listed in the
        description details for each item.
        Premium content provides portal members with access to ready-to-use
        content such as demographic and lifestyle maps as well as tools for
        geocoding, geoenrichment, network analysis, elevation analysis, and
        spatial analysis.

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            required string, username for AGOL
        ---------------     ----------------------------------------------------
        password            required string, login password for AGOL account
        ===============     ====================================================

        :Note:
          This will cost you credits.


        """
        group_id = None
        for g in self.groups:
            if g.title.lower() == "living atlas":
                group_id = g.id
                break
        params = {
            "f" : "json",
            "username" : username,
            "password" : password,
            'type' : "Premium",
            'groupId' : group_id
        }
        url = "%s/enable" % self._url
        res = self._con.post(path=url, postdata=params)
        if 'status' in res and \
           res['status'] == 'success' and \
           group_id:
            url = "%s/share" % self._url
            params = {
                "f" : "json",
                "groupId" : group_id,
                "type" : 'Premium'
            }
            res = self._con.post(path=url, postdata=params)
            return res['status'] == 'success'
        else:
            return False
        return
    #----------------------------------------------------------------------
    def enable_subscriber_atlas(self, username, password):
        """
        Enables the Premium Livinng Atlas Content for a local portal.

        Subscriber content is the collection of ready-to-use map layers,
        analytic tools, and services published by Esri that requires an
        ArcGIS Online organizational subscription account to access. This
        includes layers from Esri such as Landsat 8 imagery, NAIP imagery,
        landscape analysis layers, and historical maps. Subscriber content
        is provided as part of your organizational subscription and does
        not consume any credits. Layers included in the Living Atlas
        subscriber content are suitable for use with analysis tools.


        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        username            required string, username for AGOL
        ---------------     ----------------------------------------------------
        password            required string, login password for AGOL account
        ===============     ====================================================

        :Note:
          This will **not** cost your organization credits.


        """
        group_ids = []
        for g in self.groups:
            group_ids.append(g.id)
        params = {
            "f" : "json",
            "username" : username,
            "password" : password,
            'type' : "Subscriber"
        }
        try:
            for ids in group_ids:
                params['groupId'] = ids
                url = "%s/enable" % self._url
                res = self._con.post(path=url, postdata=params)
            for ids in group_ids:
                url = "%s/share" % self._url
                params = {
                    "f" : "json",
                    "groupId" : ids,
                    "type" : 'Subscriber'
                }
                res = self._con.post(path=url, postdata=params)
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def disable_subscriber_atlas(self):
        """
        Disables the Subscriber level Living Atlas Content for a local portal.

        """
        group_ids = []
        for g in self.groups:
            group_ids.append(g.id)
        params = {
            "f" : "json",
            'type' : "Subscriber"
        }
        try:
            for ids in group_ids:
                params['groupId'] = ids
                url = "%s/disable" % self._url
                res = self._con.post(path=url, postdata=params)
            for ids in group_ids:
                url = "%s/unshare" % self._url
                params = {
                    "f" : "json",
                    "groupId" : ids,
                    "type" : 'Subscriber'
                }
                res = self._con.post(path=url, postdata=params)
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def disable_premium_atlas(self):
        """
        Disables the Premium Livinng Atlas Content for a local portal.

        """
        group_id = None
        for g in self.groups:
            if g.title.lower() == "living atlas":
                group_id = g.id
                break
        params = {
            "f" : "json",
            "groupId" : group_id,
            'type' : "Premium"
        }
        url = "%s/disable" % self._url
        res = self._con.post(path=url, postdata=params)
        if 'status' in res and \
           res['status'] == 'success' and \
           group_id:
            url = "%s/unshare" % self._url
            params = {
                "f" : "json",
                "groupId" : group_id,
                "type" : 'Premium'
            }
            res = self._con.post(path=url, postdata=params)
            return res['status'] == 'success'
        else:
            return False
        return False

